"""
.. module:: input

:Synopsis: Input-related functions
:Author: Jesus Torrado and Antony Lewis

"""

# Global
import os
import inspect
import platform
from copy import deepcopy
from itertools import chain
from functools import reduce
from typing import Mapping, Union, Optional, TypeVar, Callable, Dict, List
from collections import defaultdict

# Local
from cobaya.conventions import products_path, kinds, separator_files, \
    get_chi2_name, get_chi2_label, Extension, FileSuffix, \
    packages_path_input
from cobaya.typing import InputDict, InfoDict, ModelDict, ExpandedParamsDict, LikesDict, \
    empty_dict
from cobaya.tools import recursive_update, str_to_list, get_base_classes, \
    fuzzy_match, deepcopy_where_possible
from cobaya.component import get_component_class, ComponentNotFoundError
from cobaya.yaml import yaml_load_file, yaml_load
from cobaya.log import LoggedError, get_logger
from cobaya.parameterization import expand_info_param
from cobaya import mpi

# Logger
logger = get_logger(__name__)


def load_input_dict(info_or_yaml_or_file: Union[InputDict, str, os.PathLike]
                    ) -> InputDict:
    if isinstance(info_or_yaml_or_file, os.PathLike):
        return load_input_file(info_or_yaml_or_file)
    elif isinstance(info_or_yaml_or_file, str):
        if "\n" in info_or_yaml_or_file:
            return yaml_load(info_or_yaml_or_file)  # type: ignore
        else:
            return load_input_file(info_or_yaml_or_file)
    elif isinstance(info_or_yaml_or_file, (dict, Mapping)):
        return deepcopy_where_possible(info_or_yaml_or_file)
    else:
        raise ValueError("The first argument must be a dictionary, file name or "
                         "yaml string with the required input options.")


def load_input_file(input_file: Union[str, os.PathLike],
                    no_mpi: bool = False,
                    help_commands: Optional[str] = None) -> InputDict:
    if no_mpi:
        mpi.set_mpi_disabled()
    input_file = str(input_file)
    stem, suffix = os.path.splitext(input_file)
    if os.path.basename(stem) in ("input", "updated"):
        raise ValueError("'input' and 'updated' are reserved file names. "
                         "Please, use a different one.")
    if suffix.lower() in Extension.yamls + (Extension.dill,):
        info = load_input_MPI(input_file)
        root, suffix = os.path.splitext(stem)
        if suffix == ".updated":
            # path may have been removed, so put in full path and name
            info["output"] = root
    else:
        # Passed an existing output_prefix?
        # First see if there is a binary info pickle
        updated_file = get_info_path(*split_prefix(input_file), ext=Extension.dill)
        if not os.path.exists(updated_file):
            # Try to find the corresponding *.updated.yaml
            updated_file = get_info_path(*split_prefix(input_file))
        try:
            info = load_input_MPI(updated_file)
        except IOError:
            err_msg = "Not a valid input file, or non-existent run to resume."
            if help_commands:
                err_msg += \
                    f" Maybe you mistyped one of the following commands: {help_commands}"
            raise ValueError(err_msg)
        # We need to update the output_prefix to resume the run *where it is*
        info["output"] = input_file
        if "post" not in info:
            # If input given this way, we obviously want to resume!
            info["resume"] = True
    return info


def load_input(input_file: str) -> InputDict:
    """
    Loads general info, and splits it into the right parts.
    """
    file_name, extension = os.path.splitext(input_file)
    file_name = os.path.basename(file_name)
    info: InputDict
    if extension.lower() in Extension.yamls:
        info = yaml_load_file(input_file) or {}  # type: ignore
    elif extension == Extension.dill:
        info = load_info_dump(input_file) or {}
    else:
        raise LoggedError(
            logger, "Extension of input file '%s' not recognized.", input_file)
    # if output_prefix not defined, default to input_file name (sans ext.) as prefix;
    if "output" not in info:
        info["output"] = file_name
    # warn if no output, since we are in shell-invocation mode.
    elif info["output"] is None:
        logger.warning("WARNING: Output explicitly suppressed with '%s: null'",
                       "output")
    # contained? Ensure that output is sent where it should
    if "CONTAINED" in os.environ:
        for out in ("output", "debug"):
            if isinstance(info.get(out), str):
                if not info[out].startswith("/"):
                    info[out] = os.path.join(products_path, info[out])
    return info


# separate MPI function, as sometimes just use load_input from root process only
@mpi.from_root
def load_input_MPI(input_file) -> InputDict:
    return load_input(input_file)


def load_info_overrides(*infos_or_yaml_or_files, **flags) -> InputDict:
    """
    Takes a number of input dictionaries (or paths to them), loads them and updates them,
    the latter ones taking precedence.

    If present, it updates the results with the kwargs if their value is not ``None``.

    Returns a deep copy of the resulting updated input dict (non-copyable object are
    retained).
    """
    info = load_input_dict(infos_or_yaml_or_files[0])  # makes deep copy if dict
    for another_info in infos_or_yaml_or_files[1:]:
        info = recursive_update(info, load_input_dict(another_info))
    for flag, value in flags.items():
        if value is not None:
            info[flag] = value
    return info


# load from dill pickle, including any lambda functions or external classes
def load_info_dump(input_file) -> InputDict:
    import dill
    with open(input_file, 'rb') as f:
        return dill.load(f)


def split_prefix(prefix):
    """
    Splits an output prefix into folder and file name prefix.

    If on Windows, allows for unix-like input.
    """
    if platform.system() == "Windows":
        prefix = prefix.replace("/", os.sep)
    folder = os.path.dirname(prefix) or "."
    file_prefix = os.path.basename(prefix)
    if file_prefix == ".":
        file_prefix = ""
    return folder, file_prefix


def get_info_path(folder, prefix, infix=None, kind="updated", ext=Extension.yamls[0]):
    """
    Gets path to info files saved by Output.
    """
    if infix is None:
        infix = ""
    elif not infix.endswith("."):
        infix += "."
    info_file_prefix = os.path.join(
        folder, prefix + (separator_files if prefix else ""))
    try:
        suffix = {"input": FileSuffix.input, "updated": FileSuffix.updated}[kind.lower()]
    except KeyError:
        raise ValueError("`kind` must be `input|updated`")
    return info_file_prefix + infix + suffix + ext


def get_used_components(*infos, return_infos=False):
    """
    Returns all requested components as a dict ``{kind: set([components])}``.
    Priors are not included.

    The list of arguments may contain base strings, which are interpreted as component
    names and added to the returned dictionary under a ``None`` key. In this case, there
    is no guarantee that the same component will not be listed both under ``None`` and
    under its particular kind.

    If ``return_infos=True`` (default: ``False``), also returns a dictionary of inputs per
    component, updated in the order in which the info arguments are given.

    Components which are just renames of others (i.e. defined with `class_name`) return
    the original class' name.
    """
    # TODO: take inheritance into account
    comps: Dict[Union[str, None], List[str]] = defaultdict(list)
    comp_infos: Dict[str, dict] = defaultdict(dict)
    for info in infos:
        if isinstance(info, str) and info not in comps[None]:
            comps[None] += [info]
            if return_infos and info not in comp_infos:
                comp_infos[info] = {}
            continue
        for kind in kinds:
            try:
                comps[kind] += [a for a in (info.get(kind) or [])
                                if a not in comps[kind]]
            except TypeError:
                raise LoggedError(
                    logger, ("Your input info is not well formatted at the '%s' block. "
                             "It must be a dictionary {'%s_i':{options}, ...}. "),
                    kind, kind)
            if return_infos:
                for c in comps[kind]:
                    comp_infos[c].update(info[kind][c] or {})
    # return dictionary of non-empty blocks
    components = {k: v for k, v in comps.items() if v}
    return (components, dict(comp_infos)) if return_infos else components


def get_default_info(component_or_class, kind=None, return_yaml=False,
                     yaml_expand_defaults=True, component_path=None,
                     input_options=empty_dict, class_name=None,
                     return_undefined_annotations=False):
    """
    Get default info for a component_or_class.
    """
    try:
        cls = get_component_class(component_or_class, kind, component_path, class_name,
                                  logger=logger)
        default_component_info = \
            cls.get_defaults(return_yaml=return_yaml,
                             yaml_expand_defaults=yaml_expand_defaults,
                             input_options=input_options)
    except ComponentNotFoundError:
        raise
    except Exception as e:
        raise LoggedError(logger,
                          "Failed to get defaults for component or class '%s' [%s]",
                          component_or_class, e)
    if return_undefined_annotations:
        annotations = {k: v for k, v in cls.get_annotations().items() if
                       k not in default_component_info}
        return default_component_info, annotations
    else:
        return default_component_info


def add_aggregated_chi2_params(param_info, all_types):
    for t in sorted(all_types):
        param_info[get_chi2_name(t)] = {"latex": get_chi2_label(t), "derived": True}


_Dict = TypeVar('_Dict', InputDict, ModelDict)


def update_info(info: _Dict, strict: bool = True, add_aggr_chi2: bool = True) -> _Dict:
    """
    Creates an updated info starting from the defaults for each component and updating it
    with the input info.
    """
    component_base_classes = get_base_classes()
    # Don't modify the original input, and convert all Mapping to consistent dict
    input_info = deepcopy_where_possible(info)
    # Creates an equivalent info using only the defaults
    updated_info: _Dict = {}
    default_params_info = {}
    default_prior_info = {}
    used_kind_members = get_used_components(input_info)
    from cobaya.component import CobayaComponent
    for block in used_kind_members:
        updated: InfoDict = {}
        updated_info[block] = updated
        input_block = input_info[block]
        name: str
        for name in used_kind_members[block]:
            # Preprocess "no options" and "external function" in input
            try:
                input_block[name] = input_block[name] or {}
            except TypeError:
                raise LoggedError(
                    logger, ("Your input info is not well formatted at the '%s' block. "
                             "It must be a dictionary {'%s_i':{options}, ...}. "),
                    block, block)
            if isinstance(name, CobayaComponent) or isinstance(name, type):
                raise LoggedError(
                    logger, ("Instances and classes should be passed a "
                             "dictionary entry of the form 'name: instance'"))
            if isinstance(input_block[name], CobayaComponent):
                logger.warning("Support for input instances is experimental")
            if isinstance(input_block[name], type) or \
                    not isinstance(input_block[name], dict):
                input_block[name] = {"external": input_block[name]}
            ext = input_block[name].get("external")
            annotations = {}
            if ext:
                if isinstance(ext, type):
                    default_class_info, annotations = \
                        get_default_info(ext, block, input_options=input_block[name],
                                         return_undefined_annotations=True)
                else:
                    default_class_info = deepcopy_where_possible(
                        component_base_classes[block].get_defaults())
            else:
                component_path = input_block[name].get("python_path")
                try:
                    default_class_info, annotations = get_default_info(
                        name, block, class_name=input_block[name].get("class"),
                        component_path=component_path, input_options=input_block[name],
                        return_undefined_annotations=True)
                except ComponentNotFoundError:
                    if strict:
                        raise
                    default_class_info, annotations = {}, {}
            updated[name] = default_class_info or {}
            # Update default options with input info
            # Consistency is checked only up to first level! (i.e. subkeys may not match)
            # Reserved attributes not necessarily already in default info:
            reserved = {"external", "class", "provides", "requires", "renames",
                        "input_params", "output_params", "python_path", "aliases",
                        "package_install"}
            options_not_recognized = set(input_block[name]).difference(
                chain(reserved, updated[name], annotations))
            if options_not_recognized:
                alternatives = {}
                available = {"external", "class", "requires", "renames"}.union(
                    updated_info[block][name])
                while options_not_recognized:
                    option = options_not_recognized.pop()
                    alternatives[option] = fuzzy_match(option, available, n=3)
                did_you_mean = ", ".join(
                    [("'%s' (did you mean %s?)" % (o, "|".join(["'%s'" % _ for _ in a]))
                      if a else "'%s'" % o)
                     for o, a in alternatives.items()])
                raise LoggedError(
                    logger, ("%s '%s' does not recognize some options: %s. "
                             "Check the documentation for '%s'."),
                    block, name, did_you_mean, block)
            updated[name].update(input_block[name])
            # save params and priors of class to combine later
            default_params_info[name] = default_class_info.get("params", {})
            default_prior_info[name] = default_class_info.get("prior", {})
    # Add priors info, after the necessary checks
    if "prior" in input_info or any(default_prior_info.values()):
        updated_info["prior"] = input_info.get("prior", {})
    for prior_info in default_prior_info.values():
        for name, prior in prior_info.items():
            if updated_info["prior"].get(name, prior) != prior:
                raise LoggedError(
                    logger, "Two different priors cannot have the same name: '%s'.", name)
            updated_info["prior"][name] = prior
    # Add parameters info, after the necessary updates and checks
    defaults_merged = merge_default_params_info(default_params_info)
    param_info: ExpandedParamsDict = merge_params_info([defaults_merged,
                                                        input_info.get("params", {})],
                                                       default_derived=False)
    updated_info["params"] = param_info  # type: ignore
    # Add aggregated chi2 params
    if info.get("likelihood") and add_aggr_chi2:
        all_types = set(chain(
            *[str_to_list(like_info.get("type", []) or [])
              for like_info in updated_info["likelihood"].values() if
              like_info is not None]))
        add_aggregated_chi2_params(param_info, all_types)
    # Add automatically-defined parameters
    if "auto_params" in updated_info:
        make_auto_params(updated_info.pop("auto_params"), param_info)
    # Add aliases for theory params (after merging!)
    for name in ("theory", "likelihood"):
        if isinstance(updated_info.get(name), dict):
            for item in updated_info[name].values():
                renames = item.get("renames")
                if renames:
                    if not isinstance(renames, Mapping):
                        raise LoggedError(
                            logger, ("'renames' should be a dictionary of name mappings "
                                     "(or you meant to use 'aliases')"))
                    renames_flat = [set([k] + str_to_list(v)) for k, v in renames.items()]
                    for p in param_info:
                        # Probably could be made faster by inverting
                        # the renames dicts *once*
                        renames_pairs = [a for a in renames_flat if p in a]
                        if renames_pairs:
                            this_renames = reduce(
                                lambda x, y: x.union(y),
                                [a for a in renames_flat if p in a])
                            param_info[p]["renames"] = \
                                list(set(chain(this_renames, str_to_list(
                                    param_info[p].get("renames", [])))).difference({p}))
    # Rest of the options
    for k, v in input_info.items():
        if k not in updated_info:
            updated_info[k] = v
    return updated_info


def merge_default_params_info(defaults: LikesDict):
    """
    Merges default parameters info for all likelihoods.
    Checks that multiple defined (=shared) parameters have equal info.
    """
    defaults_merged: LikesDict = {}
    for lik, params in defaults.items():
        for p, info in (params or {}).items():
            # if already there, check consistency
            if p in defaults_merged:
                if info != defaults_merged[p]:
                    raise LoggedError(
                        logger, ("Parameter '%s' multiply defined, but inconsistent info:"
                                 " For likelihood '%s' is '%r', but for some other "
                                 "likelihood it was '%r'. Check your defaults!"),
                        p, lik, info, defaults_merged[p])
                if mpi.is_main_process():
                    logger.debug("Parameter '%s' is multiply defined but consistent.", p)
            defaults_merged[p] = info
    return defaults_merged


def merge_params_info(params_infos, default_derived=True):
    """
    Merges parameter infos, starting from the first one
    and updating with each additional one.
    Labels (for sampled and derived) and min/max
    (just for derived params) are inherited from defaults
    (but not if one of min/max is re-defined: in that case,
    to avoid surprises, the other one is set to None=+/-inf)
    """
    current_info = {p: expand_info_param(v, default_derived) for p, v in
                    params_infos[0].items() or {}}
    for new_info in params_infos[1:]:
        if not new_info:
            continue
        for p, new_info_p in new_info.items():
            if p not in current_info:
                current_info[p] = {}
            new_info_p = expand_info_param(new_info_p)
            current_info[p].update(deepcopy(new_info_p))
            # Account for incompatibilities: "prior" and ("value" or "derived"+bounds)
            incompatibilities = {"prior": ["value", "derived", "min", "max"],
                                 "value": ["prior", "ref", "proposal"],
                                 "derived": ["prior", "drop", "ref", "proposal"]}
            for f1, incomp in incompatibilities.items():
                if f1 in new_info_p:
                    for f2 in incomp:
                        current_info[p].pop(f2, None)  # type: ignore
    # Re-sort, so that rightmost info takes precedence *also* in the sorting
    new_order_sorted = chain(*params_infos[::-1])
    # The following removes duplicates maintaining order (keeps the first occurrence)
    new_order = dict.fromkeys(new_order_sorted)
    current_info = {p: current_info[p] for p in new_order}
    return current_info


def merge_info(*infos):
    """
    Merges information dictionaries. Rightmost arguments take precedence.
    """
    assert len(infos)
    previous_info = deepcopy(infos[0])
    if len(infos) == 1:
        return previous_info
    current_info = None
    for new_info in infos[1:]:
        previous_params_info = deepcopy(previous_info.pop("params", {}) or {})
        new_params_info = deepcopy(new_info).pop("params", {}) or {}
        # NS: params have been popped, since they have their own merge function
        current_info = recursive_update(previous_info, new_info)
        current_info["params"] = merge_params_info(
            [previous_params_info, new_params_info])
        previous_info = current_info
    return current_info


def is_equal_info(info_old, info_new, strict=True, print_not_log=False, ignore_blocks=()):
    """
    Compares two information dictionaries, and old one versus a new one, and updates the
    new one for selected values of the old one.

    Set ``strict=False`` (default: ``True``) to ignore options that would not affect
    the statistics of a posterior sample, including order of params/priors/likelihoods.
    """
    myprint: Callable
    myprint_debug: Callable
    if print_not_log:
        myprint = print
        myprint_debug = lambda x: x
    else:
        myprint = logger.info
        myprint_debug = logger.debug
    myname = inspect.stack()[0][3]
    ignorable = {"debug", "resume", "force", packages_path_input,
                 "test", "version", "stop_at_error"}
    ignore = set() if strict else ignorable
    ignore = ignore.union(ignore_blocks or [])
    if set(info for info in info_old if info_old[info] is not None) - ignore \
            != set(info for info in info_new if info_new[info] is not None) - ignore:
        myprint(myname + ": different blocks or options: %r (old) vs %r (new)" % (
            set(info_old).difference(ignore), set(info_new).difference(ignore)))
        return False
    for block_name in info_old:
        if block_name in ignore or block_name not in info_new:
            continue
        block1 = deepcopy_where_possible(info_old[block_name])
        block2 = deepcopy_where_possible(info_new[block_name])
        # First, deal with root-level options (force, output, ...)
        if not isinstance(block1, dict):
            if block1 != block2:
                myprint(myname + ": different option '%s'" % block_name)
                return False
            continue
        # Now let's do components and params
        # 1. check order (it DOES matter, but just up to 1st level)
        f = list if strict else set
        if f(block1) != f(block2):
            myprint(
                myname + ": different [%s] or different order of them: %r vs %r" % (
                    block_name, list(block1), list(block2)))
            return False
        # 2. Gather general options to be ignored
        ignore_k = set()
        if not strict:
            if block_name in ["theory", "likelihood"]:
                ignore_k.update({"input_params", "output_params"})
            elif block_name == "params":
                for param in block1:
                    # Unify notation
                    block1[param] = expand_info_param(block1[param])
                    block2[param] = expand_info_param(block2[param])
                    ignore_k.update({"latex", "renames", "ref", "proposal", "min", "max"})
                    # Fixed params, it doesn't matter if they are saved as derived
                    if "value" in block1[param]:
                        block1[param].pop("derived", None)
                    if "value" in block2[param]:
                        block2[param].pop("derived", None)
                    # Renames: order does not matter
                    block1[param]["renames"] = set(block1[param].get("renames", []))
                    block2[param]["renames"] = set(block2[param].get("renames", []))
        # 3. Now check component/parameters one-by-one
        for k in block1:
            if not strict:
                # Add component-specific options to be ignored
                if block_name in kinds:
                    ignore_k_this = ignore_k.union({"python_path"})
                    if "external" not in block1[k]:
                        try:
                            component_path = block1[k].pop("python_path", None) \
                                if isinstance(block1[k], dict) else None
                            cls = get_component_class(
                                k, kind=block_name, component_path=component_path,
                                class_name=(block1[k] or {}).get("class"), logger=logger)
                            ignore_k_this.update(set(
                                getattr(cls, "_at_resume_prefer_new", {})))
                        except ImportError:
                            pass
                    # Pop ignored and kept options
                    for j in ignore_k_this:
                        block1[k].pop(j, None)
                        block2[k].pop(j, None)
            if block1[k] != block2[k]:
                # For clarity, pop common stuff before printing
                to_pop = [j for j in block1[k] if (block1[k].get(j) == block2[k].get(j))]
                [(block1[k].pop(j, None), block2[k].pop(j, None)) for j in to_pop]
                myprint(
                    myname + ": different content of [%s:%s]" % (block_name, k) +
                    " -- (re-run with `debug: True` for more info)")
                myprint_debug("%r (old) vs %r (new)" % (block1[k], block2[k]))
                return False
    return True


def get_preferred_old_values(info_old):
    """
    Returns selected values in `info_old`, which are preferred at resuming.
    """
    keep_old: InfoDict = {}
    for block_name, block in info_old.items():
        if block_name not in kinds or not block:
            continue
        for k in block:
            try:
                component_path = block[k].pop("python_path", None) \
                    if isinstance(block[k], dict) else None
                cls = get_component_class(
                    k, kind=block_name, component_path=component_path,
                    class_name=(block[k] or {}).get("class"), logger=logger)
                prefer_old_k_this = getattr(cls, "_at_resume_prefer_old", {})
                if prefer_old_k_this:
                    if block_name not in keep_old:
                        keep_old[block_name] = {}
                    keep_old[block_name].update(
                        {k: {o: block[k][o] for o in prefer_old_k_this if o in block[k]}})
            except ImportError:
                pass
    return keep_old


def make_auto_params(auto_params, params_info):
    def replace(item, tag):
        if isinstance(item, dict):
            for key, val in list(item.items()):
                item[key] = replace(val, tag)
        elif isinstance(item, str) and '%s' in item:
            item = item % tag
        return item

    for k, v in auto_params.items():
        if '%s' not in k:
            raise LoggedError(
                logger, "auto_param parameter names must have '%s' placeholder")
        replacements = v.pop('auto_range')
        if isinstance(replacements, str):
            replacements = eval(replacements)
        for value in replacements:
            params_info[k % value] = replace(deepcopy_where_possible(v), value)



import numpy as np
import pytest

from cobaya.install import install
from cobaya.model import get_model
from cobaya.yaml import yaml_load

from .common import process_packages_path

yaml = r"""
theory:
  camb:
    extra_args:
      lens_potential_accuracy: 1
likelihood:
  planck_2018_lowl.TT: null
  planck_2018_lowl.EE: null
  planck_NPIPE_highl_CamSpec.TTTEEE: null
  planckpr4lensing:
    package_install:
      github_repository: carronj/planck_PR4_lensing
      min_version: 1.0.2
params:
  tau: 0.05
  logA: 3.04920413
  As:
    value: 'lambda logA: 1e-10*np.exp(logA)'
  ns: 0.96399503
  theta_MC_100: 1.04240171
  thetastar:
    value: 'lambda theta_MC_100: 1.e-2*theta_MC_100'
    derived: false
  ombh2: 0.02235048
  omch2: 0.12121379
"""


def test_planck_NPIPE_p_CamSpec_camb_install(packages_path):
    packages_path = process_packages_path(packages_path)
    info = yaml_load(yaml)
    result = install(info, path=packages_path)
    if result is False:
        raise pytest.xfail("Some component(s) skipped")

    model = get_model(info, packages_path=packages_path)
    pars = (0.99818025, 10.35947284, 18.67072461, 7.54932654, 0.83715482,
            0.94987418, 1.23385364, 0.98781552, 1.013345)

    assert np.isclose(model.logposterior(pars).logpost, -5889.873, rtol=1e-4)



from .common import process_packages_path
from .conftest import install_test_wrapper
import os
import numpy as np
from cobaya.model import get_model
from cobaya.tools import load_module
from cobaya.component import ComponentNotInstalledError

params = {'ombh2': 0.02242, 'omch2': 0.11933, 'H0': 67.66, 'tau': 0.0561,
          'mnu': 0.06, 'nnu': 3.046, 'ns': 0.9665,
          'YHe': 0.2454, 'As': 2e-9}


def get_camb(packages_path):
    try:
        return load_module("camb", path=os.path.join(process_packages_path(packages_path),
                                                     "code", "CAMB"))
    except ModuleNotFoundError:
        raise ComponentNotInstalledError('camb')


def _get_model(packages_path, likelihood_info, skip_not_installed):
    info = {
        'params': params,
        'likelihood': {'test_likelihood': likelihood_info},
        'theory': {'camb': {'stop_at_error': True,
                            'extra_args': {'num_massive_neutrinos': 1,
                                           'halofit_version': 'mead'}}},
        'packages_path': process_packages_path(packages_path)}
    return install_test_wrapper(skip_not_installed, get_model, info)


def test_sources(packages_path, skip_not_installed):
    camb = install_test_wrapper(skip_not_installed, get_camb, packages_path)
    from camb.sources import GaussianSourceWindow

    pars = camb.set_params(**params)
    pars.set_for_lmax(500)
    pars.SourceWindows = [
        GaussianSourceWindow(redshift=0.17, source_type='counts', bias=1.2, sigma=0.04)]
    results = camb.get_results(pars)
    dic = results.get_source_cls_dict()

    # noinspection PyDefaultArgument
    def test_likelihood(_self):
        assert abs(
            _self.provider.get_source_Cl()[('source1', 'source1')][100] / dic['W1xW1'][
                100] - 1) < 0.001, \
            "CAMB gaussian source window results do not match"
        return 0

    test_likelihood_requires = {
        'source_Cl': {'sources': {'source1':
                                      {'function': 'gaussian',
                                       'source_type': 'counts',
                                       'bias': 1.2,
                                       'redshift': 0.17,
                                       'sigma': 0.04}},
                      'limber': True, 'lmax': 500}}

    model = _get_model(
        packages_path,
        {'external': test_likelihood, 'requires': test_likelihood_requires},
        skip_not_installed)
    model.loglike({})


def test_CAMBdata(packages_path, skip_not_installed):
    # noinspection PyDefaultArgument
    def test_likelihood(_self):
        return _self.provider.get_CAMBdata().tau0

    test_likelihood_requires = {'CAMBdata': None, 'Pk_grid': dict(k_max=2, z=[0, 2])}

    model = _get_model(
        packages_path,
        {'external': test_likelihood, 'requires': test_likelihood_requires},
        skip_not_installed)
    assert np.isclose(model.loglike({})[0], 14165.63, rtol=1e-4), \
        "CAMBdata object result failed"


def test_CAMB_transfer(packages_path, skip_not_installed):
    camb = install_test_wrapper(skip_not_installed, get_camb, packages_path)
    pars = camb.set_params(**{**params, 'halofit_version': 'mead'})
    pars.set_matter_power(redshifts=[0, 2], kmax=2)
    pars.WantCls = False
    results = camb.get_results(pars)
    k, z, PK1 = results.get_nonlinear_matter_power_spectrum(hubble_units=False)

    # noinspection PyDefaultArgument,PyUnresolvedReferences
    def test_likelihood(_self):
        _, _, PK = _self.provider.get_Pk_grid()
        assert np.isclose(PK[1, 30], 10294.3285)
        np.testing.assert_allclose(PK, PK1, rtol=1e-4)
        return 1

    test_likelihood_requires = {'Pk_grid': dict(k_max=2, z=[0, 2])}

    model = _get_model(
        packages_path,
        {'external': test_likelihood, 'requires': test_likelihood_requires},
        skip_not_installed)
    model.loglike()


def test_CAMB_sigma_R(packages_path, skip_not_installed):
    camb = install_test_wrapper(skip_not_installed, get_camb, packages_path)
    pars = camb.set_params(**params)
    redshifts = [0, 2, 5]
    pars.set_matter_power(redshifts=redshifts, kmax=2)
    pars.WantCls = False
    results = camb.get_results(pars)
    R = np.arange(1, 20, 1)
    sigma_R = results.get_sigmaR(R=R, hubble_units=False)[::-1, :]

    # noinspection PyDefaultArgument,PyUnresolvedReferences
    def test_likelihood(_self):
        r_out, z_out, sigma_R_out = _self.provider.get_sigma_R()
        assert np.allclose(z_out, redshifts)
        np.testing.assert_allclose(sigma_R, sigma_R_out, rtol=1e-3)
        return 1

    test_likelihood_requires = {'sigma_R': dict(z=[0, 2, 5], R=R),
                                'Pk_grid': {'k_max': 1, 'z': np.arange(0.2, 6, 1)}}

    model = _get_model(
        packages_path,
        {'external': test_likelihood, 'requires': test_likelihood_requires},
        skip_not_installed)
    model.loglike()



"""
Tests different aspects of the reparameterization layer, using an external (to force hard
parameter names) gaussian likelihood.
"""

# Global
import pytest
from scipy.stats import multivariate_normal
import numpy as np
# Local
from cobaya.yaml import yaml_load
from cobaya.run import run
from cobaya.tools import get_external_function
from cobaya.likelihood import Likelihood
from cobaya.model import get_model
from cobaya.log import LoggedError

x_func = lambda _: _ / 3
e_func = lambda _: _ + 1
b_func = "lambda a, bprime: a+2*bprime"
c_func = "lambda a, cprime: a+3*cprime"
f_func = "lambda b: b**2"
g_func = "lambda x: 3*x"
h_func = "lambda i: i**2"
j_func = "lambda b: b**2"
k_func = "lambda f: f**3"


def loglike(a, b, c, d, h, i, j):
    logp = multivariate_normal.logpdf((a, b, c, d, h, i, j), cov=0.1 * np.eye(7))
    derived = {"x": x_func(c), "e": e_func(b)}
    return logp, derived


# Info
info = {
    "likelihood":
        {"test_lik": {"external": loglike, "output_params": ["x", "e"]}},
    "sampler": {"mcmc": {"burn_in": 0, "max_samples": 10}},
    "params": yaml_load("""
       # Fixed to number
       a: 0.01
       # Fixed to function, non-explicitly requested as derived
       b: "%s"
       # Fixed to function, explicitly requested as derived
       c:
         value: "%s"
         derived: True
       # Sampled, dynamically defined from functions
       bprime:
         prior:
           min: -1
           max:  1
         drop: True
         proposal: 0.0001
       cprime:
         prior:
           min: -1
           max:  1
         drop: True
         proposal: 0.0001
       # Simple sampled parameter
       d:
         prior:
           min: -1
           max:  1
         proposal: 0.0001
       # Simple derived parameter
       e:
       # Dynamical derived parameter
       f:
         derived: "%s"
       # Dynamical derived parameter, needing non-mentioned output parameter (x)
       g:
         derived: "%s"
       # Fixing parameter whose only role is being an argument for a different one
       h: "%s"
       i: 2
       # Multi-layer: input parameter of "2nd order", i.e. dep on another dyn input
       j: "%s"
       # Multi-layer derived parameter of "2nd order", i.e. depe on another dyn derived
       k:
         derived: "%s"
    """ % (b_func, c_func, f_func, g_func, h_func, j_func, k_func))}


def test_parameterization():
    updated_info, sampler = run(info)
    products = sampler.products()
    sample = products["sample"]
    from getdist.mcsamples import MCSamplesFromCobaya
    gdsample = MCSamplesFromCobaya(updated_info, products["sample"])
    for i, point in sample:
        a = info["params"]["a"]
        b = get_external_function(info["params"]["b"])(a, point["bprime"])
        c = get_external_function(info["params"]["c"])(a, point["cprime"])
        e = get_external_function(e_func)(b)
        f = get_external_function(f_func)(b)
        g = get_external_function(info["params"]["g"]["derived"])(x_func(point["c"]))
        h = get_external_function(info["params"]["h"])(info["params"]["i"])
        j = get_external_function(info["params"]["j"])(b)
        k = get_external_function(info["params"]["k"]["derived"])(f)
        assert np.allclose(
            point[["b", "c", "e", "f", "g", "h", "j", "k"]].to_numpy(np.float64),
            [b, c, e, f, g, h, j, k])
        # Test for GetDist too (except fixed ones, ignored by GetDist)
        bcefffg_getdist = [gdsample.samples[i][gdsample.paramNames.list().index(p)]
                           for p in ["b", "c", "e", "f", "g", "j", "k"]]
        assert np.allclose(bcefffg_getdist, [b, c, e, f, g, j, k])


def test_parameterization_dependencies():
    class TestLike(Likelihood):
        params = {'a': None, 'b': None}

        def get_can_provide_params(self):
            return ['D']

        def logp(self, **params_values):
            a = params_values['a']
            b = params_values['b']
            params_values['_derived']['D'] = -7
            return a + 100 * b

    info_yaml = r"""
    params:
      aa:  
        prior: [2,4]
      bb:
        prior: [0,1]
        ref: [0.5, 0.1]
      c:
        value: "lambda aa, bb: aa+bb"  
      a: 
        value: "lambda c, aa: c*aa"  
      b: 1
      D:
      E:
       derived: "lambda D,c,a,aa: D*c/a+aa"      
    prior:
      pr: "lambda bb, a: bb-10*a"

    stop_at_error: True
    """
    test_info = yaml_load(info_yaml)
    test_info["likelihood"] = {"Like": TestLike}

    model = get_model(test_info)
    assert np.isclose(model.loglike({'bb': 0.5, 'aa': 2})[0], 105)
    assert np.isclose(model.logposterior({'bb': 0.5, 'aa': 2}).logpriors[1], -49.5)
    test_info['params']['b'] = {'value': 'lambda a, c, bb: a*c*bb'}
    like, derived = get_model(test_info).loglike({'bb': 0.5, 'aa': 2})
    assert np.isclose(like, 630)
    assert derived == [2.5, 5.0, 6.25, -7, -1.5]
    assert np.isclose(model.logposterior({'bb': 0.5, 'aa': 2}).logpriors[1], -49.5)
    test_info['params']['aa'] = 2
    test_info['params']['bb'] = 0.5
    like, derived = get_model(test_info).loglike()
    assert np.isclose(like, 630)
    assert derived == [2.5, 5.0, 6.25, -7, -1.5]

    test_info["prior"]["on_derived"] = "lambda f: 5*f"
    with pytest.raises(LoggedError) as e:
        get_model(test_info)
    assert "found and don't have a default value either" in str(e.value)

    # currently don't allow priors on derived parameters
    test_info["prior"]["on_derived"] = "lambda E: 5*E"
    with pytest.raises(LoggedError) as e:
        get_model(test_info)
    assert "that are output derived parameters" in str(e.value)



from cobaya.likelihoods.base_classes import Planck2018Clik


class TE(Planck2018Clik):
    r"""
    High-$\ell$ cross TE power spectrum \textsc{plik} likelihood of Planck's 2018 data
    release \cite{Aghanim:2019ame}.
    """
    pass



import warnings
from flaky import flaky
import numpy as np

from cobaya.run import run
from .common_sampler import body_of_sampler_test, body_of_test_speeds
from .conftest import install_test_wrapper


@flaky(max_runs=3, min_passes=1)
def test_polychord(packages_path, skip_not_installed, tmpdir):
    dimension = 3
    n_modes = 1
    info_sampler = {"polychord": {"nlive": 25 * dimension * n_modes}}
    body_of_sampler_test(info_sampler, dimension=dimension, n_modes=n_modes,
                         tmpdir=str(tmpdir),
                         packages_path=packages_path, skip_not_installed=skip_not_installed)


def test_polychord_resume(packages_path, skip_not_installed, tmpdir):
    """
    Tests correct resuming of a run, especially conserving the original blocking.

    To test preservation of the oversampling+blocking, we try to confuse the sampler by
    requesting speed measuring at resuming, and providing speeds very different from the
    real ones.
    """
    nlive = 10
    max_ndead = 2 * nlive
    dead_points = []

    def callback(sampler):
        nonlocal dead_points
        dead_points = sampler.dead[["a", "b"]].to_numpy(dtype=np.float64).copy()

    info = {
        "likelihood": {
            "A": {"external": "lambda a: stats.norm.logpdf(a)", "speed": 1},
            "B": {"external": "lambda b: stats.norm.logpdf(b)", "speed": 0.01}},
        "params": {
            "a": {"prior": {"min": 0, "max": 1}},
            "b": {"prior": {"min": 0, "max": 1}}},
        "sampler": {
            "polychord": {
                "measure_speeds": True,
                "nlive": nlive,
                "max_ndead": max_ndead,
                "callback_function": callback,
            }},
        "output": str(tmpdir)}
    install_test_wrapper(skip_not_installed, run, info)
    old_dead_points = dead_points.copy()
    info["resume"] = True
    run(info)
    assert np.allclose(old_dead_points, dead_points)


@flaky(max_runs=5, min_passes=1)
def test_polychord_multimodal(packages_path, skip_not_installed, tmpdir):
    dimension = 2
    n_modes = 2
    info_sampler = {"polychord": {"nlive": 40 * dimension * n_modes}}
    body_of_sampler_test(info_sampler, dimension=dimension, n_modes=n_modes,
                         tmpdir=str(tmpdir),
                         packages_path=packages_path, skip_not_installed=skip_not_installed)


@flaky(max_runs=3, min_passes=1)
def test_polychord_speeds(packages_path, skip_not_installed):
    info_polychord = {"polychord": {"oversample_power": 1}}
    body_of_test_speeds(info_polychord, packages_path=packages_path,
                        skip_not_installed=skip_not_installed)


@flaky(max_runs=3, min_passes=1)
def test_polychord_speeds_manual(packages_path, skip_not_installed):
    info_polychord = {"polychord": {"oversample_power": 1}}
    body_of_test_speeds(info_polychord, manual_blocking=True,
                        packages_path=packages_path,
                        skip_not_installed=skip_not_installed)


@flaky(max_runs=3, min_passes=1)
def test_polychord_unphysical(packages_path, skip_not_installed):
    """
    Tests that the effect of unphysical regions is subtracted correctly.

    To do that, it integrates a normalised 2D Guassian likelihood over a uniform prior in
    the region (-bounds, +bound) x (-bound, bound), with a x>y cut.

    The correct evidence is int(pi*L) = 1/V int(L), which for a normalised prior, that is
    1/piVol * int(L) = 1/(2*bound)**2 / 2) * 0.5.

    To get that result, we need a run with likelihood off to get the correct prior
    normalisation, which is of course the prior *density* (1/(2*bound))**2 over the result
    of the run, which is 0.5.

    We then run with the full likelihood and, as usual, divide (subtract in log) by the
    normalisation factor of the prior: the result of the prior-only run.
    """
    bound = 10
    info = {
        "likelihood": {
            "gaussian": "lambda a_0, a_1: "
                        "stats.multivariate_normal.logpdf([a_0, a_1], mean=[0,0])"},
        "prior": {"prior0": "lambda a_0, a_1: np.log(a_0 > a_1)"},
        "params": {
            "a_0": {"prior": {"min": -bound, "max": bound}},
            "a_1": {"prior": {"min": -bound, "max": bound}}},
        "sampler": {
            "polychord": {"nprior": "100nlive", "measure_speeds": False}}}
    # NB: we increase nprior wrt the default (25d=nlive) to get an accurate estimation
    #     of the unphysical region.
    info_like = info.pop("likelihood")
    info["likelihood"] = {"one": None}
    with warnings.catch_warnings():  # suppress warning log(0)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        _, sampler_prior_only = install_test_wrapper(skip_not_installed, run, info)
    logZpi = sampler_prior_only.products()["logZ"]
    logZpistd = sampler_prior_only.products()["logZstd"]
    info["likelihood"] = info_like
    with warnings.catch_warnings():  # suppress warning log(0)
        warnings.simplefilter("ignore", category=RuntimeWarning)
        _, sampler_with_like = run(info)
    logZlike = sampler_with_like.products()["logZ"]
    logZlikestd = sampler_with_like.products()["logZstd"]
    logZ = logZlike - logZpi
    sigma = logZlikestd + logZpistd
    truth = 1 / ((2 * bound) ** 2 / 2) * 0.5
    assert logZ - 2 * sigma < np.log(truth) < logZ + 2 * sigma



from .InstallableLikelihood import InstallableLikelihood
from .bao import BAO
from .DataSetLikelihood import DataSetLikelihood, _fast_chi_square
from .cmblikes import CMBlikes, make_forecast_cmb_dataset
from .des import DES
from .H0 import H0
from .Mb import Mb
from .planck_2018_CamSpec_python import Planck2018CamSpecPython
from .planck_clik import PlanckClik, Planck2018Clik, Planck2015Clik
from .planck_pliklite import PlanckPlikLite
from .sn import SN

_is_abstract = True



"""
.. module:: log

:Synopsis: Manages logging and error handling
:Author: Jesus Torrado

"""

# Global
import os
import sys
import logging
import platform
import traceback
from copy import deepcopy
import functools
from random import shuffle, choice

# Local
from cobaya import mpi


class LoggedError(Exception):
    """
    Dummy exception, to be raised when the originating exception
    has been cleanly handled and logged.

    Prints the error message even if caught.
    """

    def __init__(self, logger, *args, **kwargs):
        if isinstance(logger, str):
            logger = get_logger(logger)
        if not isinstance(logger, logging.Logger):
            raise SyntaxError("The first argument of %s must be a logger "
                              "instance or name." %
                              self.__class__.__name__)
        if args:
            # If the exception is going to be caught, we may not want to print the msg
            # at logger.error level, but e.g. debug level.
            level = kwargs.pop("level", "error") or "error"
            getattr(logger, level)(*args, **kwargs)
        msg = args[0] if len(args) else ""
        if msg and len(args) > 1:
            msg = msg % args[1:]
        super().__init__(msg)


# Exceptions that will never be ignored when a component's calculation fails
always_stop_exceptions = (LoggedError, KeyboardInterrupt, SystemExit, NameError,
                          SyntaxError, AttributeError, KeyError, ImportError, TypeError)


def is_debug(log=None):
    log = log or logging.root
    return log.getEffectiveLevel() <= logging.DEBUG


def get_logger(name):
    if name.startswith('cobaya.'):
        name = name.split('.')[-1]
    return logging.getLogger(add_color_to_name(name))


# Some legible color combinations
color_strs = {
    "red_bold": "\x1b[31;1m",
    "green_bold": "\x1b[32;1m",
    "yellow_bold": "\x1b[33;1m",
    "blue_bold": "\x1b[34;1m",
    "magenta_bold": "\x1b[35;1m",
    "cyan_bold": "\x1b[36;1m",
    "light_red_bold": "\x1b[91;1m",
    "light_green_bold": "\x1b[92;1m",
    "light_yellow_bold": "\x1b[93;1m",
    "light_blue_bold": "\x1b[94;1m",
    "light_magenta_bold": "\x1b[95;1m",
    "light_cyan_bold": "\x1b[96;1m",
    # With background
    "light_grey_on_red_bold": "\x1b[37;1;41m",
    "light_grey_on_green_bold": "\x1b[37;1;42m",
    "light_grey_on_yellow_bold": "\x1b[37;1;43m",
    "light_grey_on_blue_bold": "\x1b[37;1;44m",
    "light_grey_on_magenta_bold": "\x1b[37;1;45m",
    "light_grey_on_cyan_bold": "\x1b[37;1;46m",
    "blue_on_light_green_bold": "\x1b[34;1;102m",
    "blue_on_light_yellow_bold": "\x1b[34;1;103m",
    "light_yellow_on_blue_bold": "\x1b[93;1;44m",
    "blue_on_light_cyan": "\x1b[34;1;106m",
    "red_on_white_bold": "\x1b[31;1;107m",
    "blue_on_white_bold": "\x1b[34;1;107m",
    "magenta_on_white_bold": "\x1b[35;1;107m",
}

reset_str = "\x1b[0m"

current_color_pool = []


def add_color_to_name(name, color=None):
    if not os.getenv('COBAYA_COLOR'):
        return name
    # TODO: implement for Windows, see
    # https://docs.microsoft.com/en-us/windows/console/console-virtual-terminal-sequences
    if platform.system() == "Windows":
        return name
    if color is None:
        # Choose one at random (ensure no repetition, unless too many requested)
        global current_color_pool
        if not current_color_pool:
            current_color_pool = list(color_strs.values())
            shuffle(current_color_pool)
        i = choice(list(range(len(current_color_pool))))
        color_str = current_color_pool.pop(i)
    else:
        color_str = color_strs.get(color)
    return color_str + name + reset_str


def abstract(method):
    # abstract method decorator for base class HasLogger methods

    # If an @abstract method is called dynamically from another function,
    # you get a logged error that it's not implemented.
    # An @abstract methods also will not be picked up by the dependency analyser, so
    # a class with only an @abstract method implementation of X will not be assigned to
    # provide X. Descendants can of course override @abstract methods to implement them.

    @functools.wraps(method)
    def not_implemented(self, *args, **kwargs):
        if getattr(getattr(self, method.__name__, None), '_is_abstract', None):
            # OK to call if called via super, but not if not over-ridden
            raise LoggedError(self.log, "%s NotImplemented in %s", method.__name__,
                              self.__class__.__name__)
        else:
            return method(self, *args, **kwargs)

    not_implemented._is_abstract = True  # type: ignore

    return not_implemented


class NoLogging:
    def __init__(self, level=logging.WARNING):
        self._level = level

    def __enter__(self):
        if self._level:
            logging.disable(self._level)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        if self._level:
            logging.disable(logging.NOTSET)


def exception_handler(exception_type, exception_instance, trace_back):
    # Do not print traceback if the exception has been handled and logged by LoggedError
    # MPI abort if all processes don't raise exception in short timeframe (e.g. deadlock).
    want_abort = not mpi.time_out_barrier()
    _logger_name = "exception handler"
    log = logging.getLogger(_logger_name)

    if exception_type == LoggedError:
        # make show error easily visible at end of log
        if mpi.more_than_one_process():
            log.error(str(exception_instance))
        if want_abort:
            mpi.abort_if_mpi()
        if is_debug(log):
            return  # no traceback printed
    elif issubclass(exception_type, mpi.OtherProcessError):
        log.info(str(exception_instance))
        if is_debug(log):
            return  # no traceback printed

    line = "-------------------------------------------------------------\n"
    log.critical(line[len(_logger_name) + 5:] + "\n" +
                 "".join(traceback.format_exception(
                     exception_type, exception_instance, trace_back)) +
                 line)
    if exception_type == KeyboardInterrupt:
        log.critical("Interrupted by the user.")
    elif is_debug(log):
        log.critical(
            "Some unexpected ERROR occurred. "
            "You can see the exception information above.\n"
            "We recommend trying to reproduce this error with '%s:True' in the input.\n"
            "If you cannot solve it yourself and need to report it, "
            "include the debug output,\n"
            "which you can send it to a file setting '%s:[some_file_name]'.",
            "debug", "debug")
    # Exit all MPI processes
    if want_abort:
        mpi.abort_if_mpi()


def logger_setup(debug=None, debug_file=None):
    """
    Configuring the root logger, for its children to inherit level, format and handlers.

    Level: if debug=True, take DEBUG. If numerical, use ""logging""'s corresponding level.
    If string, set debug level and use if as ``debug_file`` (unless specified separately).
    Default: INFO
    """
    if debug is True or os.getenv('COBAYA_DEBUG'):
        level = logging.DEBUG
    elif debug in (False, None):
        level = logging.INFO
    elif isinstance(debug, int):
        level = debug
    elif isinstance(debug, str):
        level = logging.DEBUG
        debug_file = debug_file or debug
    else:
        raise ValueError(
            f"Bad value for debug: {debug}. Set to bool|str(file)|int(level).")
    # Set the default level, to make sure the handlers have a higher one
    logging.root.setLevel(level)
    debug = is_debug(logging.root)

    # Custom formatter
    class MyFormatter(logging.Formatter):
        def format(self, record):
            fmt = ((" %(asctime)s " if debug else "") +
                   "[" + ("%d : " % mpi.get_mpi_rank()
                          if mpi.more_than_one_process() else "") +
                   "%(name)s" + "] " +
                   {logging.ERROR: "*ERROR* ",
                    logging.WARNING: "*WARNING* "}.get(record.levelno, "") +
                   "%(message)s")
            self._style._fmt = fmt
            return super().format(record)

    # Configure stdout handler
    handle_stdout = logging.StreamHandler(sys.stdout)
    handle_stdout.setLevel(level)
    handle_stdout.setFormatter(MyFormatter())
    # log file? Create and reduce stdout level to INFO
    if debug_file:
        file_stdout = logging.FileHandler(debug_file, mode="w")
        file_stdout.setLevel(level)
        handle_stdout.setLevel(logging.INFO)
        file_stdout.setFormatter(MyFormatter())
        logging.root.addHandler(file_stdout)
    # Add stdout handler only once!
    # noinspection PyUnresolvedReferences
    try:
        stdout_handler = next(
            h for h in logging.root.handlers if getattr(h, "stream", None) == sys.stdout)
        # If there is one, update it's logging level and formatter
        stdout_handler.setLevel(handle_stdout.level)
    except StopIteration:
        # If there is none, add it!
        logging.root.addHandler(handle_stdout)
    # Configure the logger to manage exceptions
    sys.excepthook = exception_handler


def get_traceback_text(exec_info):
    return "".join(["-"] * 20 + ["\n\n"] +
                   list(traceback.format_exception(*exec_info)) +
                   ["\n"] + ["-"] * 37)


class HasLogger:
    """
    Class having a logger with its name (or an alternative one).

    Has magic methods to ignore the logger at (de)serialization.
    """

    def set_logger(self, lowercase=True, name=None):
        name = name or self.__class__.__name__
        if lowercase:
            name = name.lower()
        self.log = logging.getLogger(add_color_to_name(name))

    # Copying and pickling
    def __deepcopy__(self, memo=None):
        new = (lambda cls: cls.__new__(cls))(self.__class__)
        new.__dict__ = {k: deepcopy(v) for k, v in self.__dict__.items() if k != "log"}
        return new

    def __getstate__(self):
        return deepcopy(self).__dict__

    def __setstate__(self, d):
        self.__dict__ = d
        self.set_logger()

    def is_debug(self):
        return is_debug(self.log)

    def is_debug_and_mpi_root(self):
        return is_debug(self.log) and mpi.is_main_process()

    @mpi.root_only
    def mpi_warning(self, msg, *args, **kwargs):
        self.log.warning(msg, *args, **kwargs)

    @mpi.root_only
    def mpi_info(self, msg, *args, **kwargs):
        self.log.info(msg, *args, **kwargs)

    @mpi.root_only
    def mpi_debug(self, msg, *args, **kwargs):
        self.log.debug(msg, *args, **kwargs)



from cobaya.likelihoods.base_classes import Planck2018CamSpecPython


class TT(Planck2018CamSpecPython):
    r"""
    High-$\ell$ temperature-only \textsc{CamSpec} likelihood of Planck's 2018 data release
    \cite{Aghanim:2019ame} (native Python re-implemetation, by A.~Lewis).
    """
    pass



r"""
.. module:: planck_clik

:Synopsis: Definition of the clik-based likelihoods
:Author: Jesus Torrado (initially based on MontePython's version
         by Julien Lesgourgues and Benjamin Audren)

"""
# Global
import os
import sys
import numpy as np
from packaging import version

# Local
from cobaya.likelihood import Likelihood
from cobaya.log import LoggedError, get_logger
from cobaya.input import get_default_info
from cobaya.install import pip_install, download_file
from cobaya.component import ComponentNotInstalledError, load_external_module
from cobaya.tools import are_different_params_lists, create_banner, VersionCheckError

_deprecation_msg_2015 = create_banner("""
The likelihoods from the Planck 2015 data release have been superseded
by the 2018 ones, and will eventually be deprecated.
""")

clik_url = 'https://github.com/benabed/clik/archive/refs/heads/main.zip'
pla_url_prefix = r"https://pla.esac.esa.int/pla-sl/data-action?COSMOLOGY.COSMOLOGY_OID="

last_version_supp_data_and_covmats = "v2.01"
last_version_clik = "16.0"
min_version_clik = "3.1"


class PlanckClik(Likelihood):
    # Data type for aggregated chi2 (case sensitive)
    type = "CMB"

    path: str
    clik_file: str

    def initialize(self):
        if "2015" in self.get_name():
            for line in _deprecation_msg_2015.split("\n"):
                self.log.warning(line)
        try:
            install_path = (
                lambda p: self.get_code_path(p) if p else None)(self.packages_path)
            # min_version here is checked inside get_clik_import_path, since it is
            # displayed in the folder name and cannot be retrieved from the module.
            clik = load_clik(
                "clik", path=self.path, install_path=install_path,
                get_import_path=get_clik_import_path, logger=self.log,
                not_installed_level="debug")
        except VersionCheckError as excpt:
            raise VersionCheckError(
                str(excpt) + " Upgrade with `cobaya-install planck_2018_lowl.TT "
                             "--upgrade`.")
        except ComponentNotInstalledError as excpt:
            raise ComponentNotInstalledError(
                self.log, (f"Could not find clik: {excpt}. "
                           "To install it, run `cobaya-install planck_2018_lowl.TT`"))
        # Loading the likelihood data
        data_path = get_data_path(self.__class__.get_qualified_class_name())
        if not os.path.isabs(self.clik_file):
            self.path_data = getattr(self, "path_data", os.path.join(
                self.path or self.packages_path, "data", data_path))
            self.clik_file = os.path.join(self.path_data, self.clik_file)
        # Differences in the wrapper for lensing and non-lensing likes
        self.lensing = clik.try_lensing(self.clik_file)
        try:
            self.clik = clik.clik_lensing(self.clik_file) if self.lensing \
                else clik.clik(self.clik_file)
        except clik.lkl.CError:
            # Is it that the file was not found?
            if not os.path.exists(self.clik_file):
                raise ComponentNotInstalledError(
                    self.log, "The .clik file was not found where specified in the "
                              "'clik_file' field of the settings of this likelihood. "
                              "Maybe the 'path' given is not correct? The full path where"
                              " the .clik file was searched for is '%s'", self.clik_file)
            # Else: unknown clik error
            self.log.error("An unexpected error occurred in clik (possibly related to "
                           "multiple simultaneous initialization, or simultaneous "
                           "initialization of incompatible likelihoods; e.g. polarised "
                           "vs non-polarised 'lite' likelihoods. See error info below:")
            raise
        self.l_maxs = self.clik.get_lmax()
        # calculate requirements here so class can also be separately instantiated
        requested_cls = ["tt", "ee", "bb", "te", "tb", "eb"]
        if self.lensing:
            has_cl = [lmax != -1 for lmax in self.l_maxs]
            requested_cls = ["pp"] + requested_cls
        else:
            has_cl = self.clik.get_has_cl()
        self.requested_cls = [cl for cl, i in zip(requested_cls, has_cl) if int(i)]
        self.l_maxs_cls = [lmax for lmax, i in zip(self.l_maxs, has_cl) if int(i)]
        self.expected_params = list(self.clik.extra_parameter_names)
        # Placeholder for vector passed to clik
        length = (len(self.l_maxs) if self.lensing else len(self.clik.get_has_cl()))
        self.vector = np.zeros(np.sum(self.l_maxs) + length + len(self.expected_params))

    def initialize_with_params(self):
        # Check that the parameters are the right ones
        differences = are_different_params_lists(
            self.input_params, self.expected_params, name_A="given", name_B="expected")
        if differences:
            raise LoggedError(
                self.log, "Configuration error in parameters: %r. "
                          "If this has happened without you fiddling with the defaults, "
                          "please open an issue in GitHub.", differences)

    def get_requirements(self):
        # State requisites to the theory code
        return {'Cl': dict(zip(self.requested_cls, self.l_maxs_cls))}

    def logp(self, **params_values):
        # get Cl's from the theory code
        cl = self.provider.get_Cl(units="FIRASmuK2")
        return self.log_likelihood(cl, **params_values)

    def log_likelihood(self, cl, **params_values):
        # fill with Cl's
        self.vector[:-len(self.expected_params)] = np.concatenate(
            [(cl[spectrum][:1 + lmax] if spectrum not in ["tb", "eb"]
              else np.zeros(1 + lmax))
             for spectrum, lmax in zip(self.requested_cls, self.l_maxs_cls)])
        # check for nan's: mey produce a segfault in clik
        # dot product is apparently the fastest way in threading-enabled numpy
        if np.isnan(np.dot(self.vector, self.vector)):
            return -np.inf
        # fill with likelihood parameters
        self.vector[-len(self.expected_params):] = (
            [params_values[p] for p in self.expected_params])
        loglike = self.clik(self.vector)[0]
        # "zero" of clik, and sometimes nan's returned
        if np.allclose(loglike, -1e30) or np.isnan(loglike):
            loglike = -np.inf
        return loglike

    def close(self):
        del self.clik  # MANDATORY: forces deallocation of the Cython class
        # Actually, it does not work for low-l likelihoods, which is quite dangerous!

    @classmethod
    def get_code_path(cls, path):
        return os.path.realpath(os.path.join(path, "code", common_path))

    @classmethod
    def is_compatible(cls):
        import platform
        return platform.system() != "Windows"

    @classmethod
    def is_installed(cls, reload=False, **kwargs):
        code_path = common_path
        data_path = get_data_path(cls.get_qualified_class_name())
        result = True
        if kwargs.get("code", True):
            result &= bool(is_installed_clik(
                os.path.realpath(os.path.join(kwargs["path"], "code", code_path)),
                reload=reload))
        if kwargs.get("data", True):
            # NB: will never raise VersionCheckerror, since version number is in the path
            _, filename = get_product_id_and_clik_file(cls.get_qualified_class_name())
            result &= os.path.exists(os.path.realpath(
                os.path.join(kwargs["path"], "data", data_path, filename)))
            # Check for additional data and covmats -- can raise VersionCheckerror
            from cobaya.likelihoods.planck_2018_lensing import native
            result &= native.is_installed(**kwargs)
        return result

    @classmethod
    def install(cls, path=None, force=False, code=True, data=True,
                no_progress_bars=False, **_kwargs):
        name = cls.get_qualified_class_name()
        log = get_logger(name)
        path_names = {"code": common_path, "data": get_data_path(name)}
        import platform
        if platform.system() == "Windows":
            log.error("Not compatible with Windows.")
            return False
        global _clik_install_failed
        if _clik_install_failed:
            log.info("Previous clik install failed, skipping")
            return False
        # Create common folders: all planck likelihoods share install
        # folder for code and data
        paths = {}
        for s in ("code", "data"):
            if eval(s):
                paths[s] = os.path.realpath(os.path.join(path, s, path_names[s]))
                if not os.path.exists(paths[s]):
                    os.makedirs(paths[s])
        success = True
        # Install clik
        if code and (not is_installed_clik(paths["code"]) or force):
            log.info("Installing the clik code.")
            success *= install_clik(paths["code"], no_progress_bars=no_progress_bars)
            if not success:
                log.warning("clik code installation failed! "
                            "Try configuring+compiling by hand at " + paths["code"])
                _clik_install_failed = True
        if data:
            # 2nd test, in case the code wasn't there but the data is:
            if force or not cls.is_installed(path=path, code=False, data=True):
                log.info("Downloading the likelihood data.")
                product_id, _ = get_product_id_and_clik_file(name)
                # Download and decompress the particular likelihood
                url = pla_url_prefix + product_id
                # Helper for the progress bars: some known product download sizes
                # (no actual effect if missing or wrong!)
                size = {"1900": 314153370, "1903": 4509715660, "151902": 60293120,
                        "151905": 5476083302, "151903": 8160437862}.get(product_id)
                if not download_file(url, paths["data"], size=size,
                                     logger=log, no_progress_bars=no_progress_bars):
                    log.error("Not possible to download this likelihood.")
                    success = False
                # Additional data and covmats, stored in same repo as the
                # 2018 python lensing likelihood
                from cobaya.likelihoods.planck_2018_lensing import native
                if not native.is_installed(data=True, path=path):
                    success *= native.install(path=path, force=force, code=code,
                                              data=data,
                                              no_progress_bars=no_progress_bars)
        return success


# Installation routines ##################################################################

# path to be shared by all Planck likelihoods
common_path = "planck"

# To see full clik build output even if installs OK (e.g. to check warnings)
_clik_verbose = any((s in os.getenv('TRAVIS_COMMIT_MESSAGE', ''))
                    for s in ["clik", "planck"])
# Don't try again to install clik if it failed for a previous likelihood
_clik_install_failed = False


def get_data_path(name):
    return common_path + "_%s" % get_release(name)


def get_release(name):
    return next(re for re in ["2015", "2018"] if re in name)


def get_clik_source_folder(starting_path):
    """
    Starting from the installation folder, returns the subdirectory from which the
    compilation must be run.

    In practice, crawls inside the install folder ``packages/code/planck``
    until >1 subfolders.

    Raises ``FileNotFoundError`` if no clik install was found.
    """
    source_dir = starting_path
    while len(folders := [f for f in os.listdir(source_dir)
                          if os.path.isdir(os.path.join(source_dir, f))]) <= 1:
        if len(folders) == 0:
            raise FileNotFoundError(
                f"Could not find a clik installation under {starting_path}")
        source_dir = os.path.join(source_dir, folders[0])
    return source_dir


def get_clik_import_path(path, min_version=min_version_clik):
    """
    Starting from the installation folder, returns the subdirectory from which the
    ``clik`` module must be imported.

    Raises ``FileNotFoundError`` if no clik install was found, or
    :class:`tools.VersionCheckError` if the installed version is too old.
    """
    clik_src_path = get_clik_source_folder(path)
    version_file = os.path.join(clik_src_path, 'svnversion')
    if os.path.exists(version_file):
        with open(version_file, 'r') as f:
            installed_version = version.parse(f.readline().split("_")[-1].split('-')[0])
    else:
        installed_version = version.parse(clik_src_path.rstrip(os.sep).split("-")[-1])
    if installed_version < version.parse(min_version):
        raise VersionCheckError(
            f"Installed version of the Plack likelihood code 'clik' ({installed_version})"
            f" older than minimum required one ({last_version_clik}).")
    elif installed_version > version.parse(last_version_clik):
        raise ValueError("This should not happen: min version needs update.")
    return os.path.join(clik_src_path, 'lib/python/site-packages')


def load_clik(*args, **kwargs):
    """
    Just a wrapper around :func:`component.load_external_module`, that checks that we are
    not being fooled by the wrong `clik <https://pypi.org/project/click/>`_.
    """
    clik = load_external_module(*args, **kwargs)
    if not hasattr(clik, "try_lensing"):
        raise ComponentNotInstalledError(
            kwargs.get("logger"), "Loaded wrong clik: `https://pypi.org/project/click/`")
    return clik


def is_installed_clik(path, reload=False):
    # min_version here is checked inside get_clik_import_path, since it is displayed
    # in the folder name and cannot be retrieved from the module.
    try:
        return bool(load_clik(
            "clik", path=path, get_import_path=get_clik_import_path,
            reload=reload, logger=get_logger("clik"), not_installed_level="debug"))
    except ComponentNotInstalledError:
        return False


def execute(command):
    from subprocess import Popen, PIPE, STDOUT
    if _clik_verbose:
        process = Popen(command, stdout=PIPE, stderr=STDOUT)
        out = []
        assert process.stdout
        while (nextline := process.stdout.readline()) != b"" or process.poll() is None:
            sys.stdout.buffer.write(nextline)
            out.append(nextline)
            sys.stdout.flush()
        _, err = process.communicate()
        return b"finished successfully" in out[-1]
    else:
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
        OK = b"finished successfully" in stdout.split(b"\n")[-2]
        if not OK:
            print(stdout.decode('utf-8'))
            print(stderr.decode('utf-8'))
        return OK


def install_clik(path, no_progress_bars=False):
    log = get_logger("clik")
    log.info("Installing pre-requisites...")
    for req in ("cython", "astropy"):
        exit_status = pip_install(req)
        if exit_status:
            raise LoggedError(log, "Failed installing '%s'.", req)
    log.info("Downloading...")
    if not download_file(clik_url, path, no_progress_bars=no_progress_bars, logger=log):
        log.error("Not possible to download clik.")
        return False
    source_dir = get_clik_source_folder(path)
    log.info('Installing from directory %s' % source_dir)
    cwd = os.getcwd()
    try:
        os.chdir(source_dir)
        log.info("Configuring... (and maybe installing dependencies...)")
        flags = ["--install_all_deps",
                 "--extra_lib=m"]  # missing for some reason in some systems, but harmless
        if not execute([sys.executable, "waf", "configure"] + flags):
            log.error("Configuration failed!")
            return False
        log.info("Compiling...")
        if not execute([sys.executable, "waf", "install"]):
            log.error("Compilation failed!")
            return False
    finally:
        os.chdir(cwd)
    log.info("Finished!")
    return True


def get_product_id_and_clik_file(name):
    """Gets the PLA product info from the defaults file."""
    defaults = get_default_info(name, "likelihood")
    return defaults.get("product_id"), defaults.get("clik_file")


class Planck2015Clik(PlanckClik):
    bibtex_file = 'planck2015.bibtex'


class Planck2018Clik(PlanckClik):
    bibtex_file = 'planck2018.bibtex'



from cobaya.likelihoods.base_classes import CMBlikes
from cobaya.likelihoods.base_classes.planck_clik import last_version_supp_data_and_covmats


class native(CMBlikes):
    r"""
    Lensing likelihood of Planck's 2018 data release based on temperature+polarization
    map-based lensing reconstruction \cite{Aghanim:2018oex} (native Python
    re-implementation by A.~Lewis).
    """
    install_options = {"github_repository": "CobayaSampler/planck_supp_data_and_covmats",
                       "github_release": last_version_supp_data_and_covmats}

    bibtex_file = 'PlanckLensing2018.bibtex'


class CMBMarged(native):
    r"""
    Lensing likelihood of Planck's 2018 data release based on temperature+polarization
    map-based lensing reconstruction, marginalized over the CMB power spectra
    \cite{Aghanim:2018oex} (native Python re-implementation by A.~Lewis).
    """
    pass



"""
.. module:: collection

:Synopsis: Classes to store the Montecarlo samples and single points.
:Author: Jesus Torrado and Antony Lewis

"""

# Global
import os
import functools
import warnings
from copy import deepcopy
from typing import Union, Sequence, Optional, Tuple, Iterable
from numbers import Number
import numpy as np
import pandas as pd
from getdist import MCSamples, chains  # type: ignore
from getdist.chains import WeightedSamples, WeightedSampleError  # type: ignore

# Local
from cobaya.conventions import OutPar, minuslogprior_names, chi2_names, \
    derived_par_name_separator, minuslogprior_labels, chi2_labels, minuslogpost_label
from cobaya.tools import load_DataFrame
from cobaya.log import LoggedError, HasLogger, NoLogging
from cobaya.model import Model, LogPosterior, DummyModel

# Suppress getdist output
chains.print_load_details = False

# Size of fast numpy cache
# (used to avoid "setting" in Pandas too often, which is expensive)
_default_cache_size = 200

# Sample types, for the purposes of e.g. knowing whether skip/thin operations are allowed.
sample_types = ["mcmc", "nested"]


def check_index(i, imax):
    """Makes sure that we don't reach the empty part of the dataframe."""
    if (i > 0 and i >= imax) or (i < 0 and -i > imax):
        raise IndexError("Trying to access a sample index larger than "
                         f"the amount of samples ({imax})!")
    if i < 0:
        return imax + i
    return i


def check_slice(ij: slice, imax=None):
    """
    Restricts a slice to the non-empty part of the DataFrame.

    Notice that slices are never supposed to raise IndexError, but an empty list at worst!
    """
    newlims = {"start": ij.start, "stop": ij.stop}
    if ij.start is None:
        newlims["start"] = 0
    if ij.stop is None and imax is not None:
        newlims["stop"] = imax
    if imax is not None:
        for limname, lim in newlims.items():
            if lim >= 0:
                newlims[limname] = min(imax, lim)
            else:
                newlims[limname] = imax + lim
    return slice(newlims["start"], newlims["stop"], ij.step)


def apply_temperature(logpost, temperature):
    """Applies sampling temperature to a log-posteior."""
    return logpost / temperature


def remove_temperature(logpost, temperature):
    """Removes sampling temperature from a log-posteior."""
    return apply_temperature(logpost, 1 / temperature)


def apply_temperature_cov(cov, temperature):
    """
    Convert covariance matrix to that of ``probability^(1/temperature)`` posterior.
    """
    return cov * temperature


def remove_temperature_cov(cov, temperature):
    """
    Convert covariance matrix of ``probability^(1/temperature)`` posterior to original
    one.
    """
    return cov / temperature


def compute_temperature(logpost, logprior, loglike, check=True):
    """
    Returns the temperature of a sample.

    If ``check=True`` and the log-probabilites passed are arrays, checks consistency
    of the sample temperature, and raises ``AssertionError`` if inconsistent.
    """
    temp = (logprior + loglike) / logpost
    if check and isinstance(temp, Iterable) and len(temp) > 1:
        assert np.allclose(temp, temp[0]), "Inconsistent temperature in sample."
        temp = temp[0]
    return temp


def detempering_weights_factor(tempered_logpost, temperature):
    """
    Returns the detempering factors for the weights of a tempered sample, i.e. if ``w_t``
    is the weight of the tempered sample, then the weight of the unit-temperature one is
    ``w_t * f``, where the ``f`` returned by this method is
    ``exp(logp * (1 - 1/temperature))``, where ``logp`` is the (untempered) logposterior.

    Factors are normalized so that the largest equals one.
    """
    if temperature == 1:
        return np.ones(np.atleast_1d(tempered_logpost).shape)
    log_ratio = remove_temperature(tempered_logpost, temperature) - tempered_logpost
    return np.exp(log_ratio - max(log_ratio))


class BaseCollection(HasLogger):
    def __init__(self, model, name=None, temperature=None):
        self.name = name
        self.set_logger(name)
        self.sampled_params = list(model.parameterization.sampled_params())
        self.derived_params = list(model.parameterization.derived_params())
        self.minuslogprior_names = minuslogprior_names(model.prior)
        self.chi2_names = chi2_names(model.likelihood)
        columns = [OutPar.weight, OutPar.minuslogpost]
        columns += list(self.sampled_params)
        # Just in case: ignore derived names as likelihoods: would be duplicate cols
        columns += [p for p in self.derived_params if p not in self.chi2_names]
        columns += [OutPar.minuslogprior] + self.minuslogprior_names
        columns += [OutPar.chi2] + self.chi2_names
        self.temperature = temperature if temperature is not None else 1
        self.columns = columns
        self._cache_aux_model_quantities(model)

    def _cache_aux_model_quantities(self, model):
        """
        Stores some auxiliary Model-related variables to allow e.g. for interfacing other
        codes without needing to use the Model again.

        Can be called inside these interfaces in case the model has changed.
        """
        self._cached_labels = deepcopy(model.parameterization.labels())
        self._cached_labels[OutPar.minuslogpost] = minuslogpost_label()
        self._cached_labels.update(minuslogprior_labels(model.prior))
        self._cached_labels.update(chi2_labels(model.likelihood))
        self._cached_renames = deepcopy(model.parameterization.sampled_params_renames())
        self._cached_ranges = None
        if not isinstance(model, DummyModel):  # can happen during post-processing
            self._cached_ranges = dict(zip(
                self.sampled_params,
                model.prior.bounds(confidence_for_unbounded=0.9999995)))  # 5 sigmas
        for p, p_info in model.parameterization.derived_params_info().items():
            mini, maxi = p_info.get("min", -np.inf), p_info.get("max", np.inf)
            some_bound_specified = np.isfinite(mini) or np.isfinite(maxi)
            if some_bound_specified:
                self._cached_ranges[p] = [mini, maxi]


def ensure_cache_dumped(method):
    """
    Decorator for SampleCollection methods that need the cache cleaned before running.
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        self._cache_dump()  # pylint: disable=protected-access
        return method(self, *args, **kwargs)

    return wrapper


class SampleCollection(BaseCollection):
    """
    Holds a collection of samples, stored internally into a ``pandas.DataFrame``.

    The DataFrame itself is accessible as the ``SampleCollection.data`` property,
    but slicing can be done on the ``SampleCollection`` itself
    (returns a copy, not a view).

    When ``temperature`` is different from 1, weights and log-posterior (but not priors
    or likelihoods' chi squared's) are those of the tempered sample, obtained assuming a
    posterior raised to the power of ``1/temperature``. Functions returning statistics,
    e.g. :func:`~SampleCollection.cov`, will return the statistics of the original
    (untempered) posterior, unless indicated otherwise with a keyword argument.

    Note for developers: when expanding this class or inheriting from it, always access
    the underlying DataFrame as ``self.data`` and not ``self._data``, to ensure the cache
    has been dumped. If you really need to access the actual attribute ``self._data`` in a
    method, make sure to decorate it with ``@ensure_cache_dumped``.
    """

    def __init__(self, model, output=None, cache_size=_default_cache_size, name=None,
                 extension=None, file_name=None, resuming=False, load=False,
                 temperature=None, onload_skip=0, onload_thin=1, sample_type=None):
        super().__init__(model, name)
        if sample_type is not None and (not isinstance(sample_type, str) or
                                        not sample_type.lower() in sample_types):
            raise LoggedError(self.log, "'sample_type' must be one of %r.", sample_types)
        self.sample_type = sample_type.lower() if sample_type is not None else sample_type
        self.cache_size = cache_size
        self._data = None
        self._n = None
        # Create/load the main data frame and the tracking indices
        # Create the DataFrame structure
        if output:
            if file_name:
                self.file_name = file_name
                self.driver = output.kind
            else:
                self.file_name, self.driver = output.prepare_collection(
                    name=self.name, extension=extension)
            self.root_file_name = os.path.join(output.folder, output.prefix)
        else:
            self.driver = "dummy"
            self.root_file_name = None
        if resuming or load:
            if output:
                try:
                    self._out_load(skip=onload_skip)
                    if onload_thin != 1:
                        self.thin_samples(onload_thin, inplace=True)
                    if load:
                        self.columns = list(self.data.columns)
                        loaded_chi2_names = set(
                            name for name in self.columns
                            if name.startswith(OutPar.chi2 + derived_par_name_separator))
                        loaded_chi2_names.discard(
                            OutPar.chi2 + derived_par_name_separator + 'prior')
                        if set(self.chi2_names).difference(loaded_chi2_names):
                            raise LoggedError(self.log,
                                              "Input samples do not have chi2 values "
                                              "matching likelihoods in the model:\n "
                                              "found: %s\nexpected: %s\n",
                                              loaded_chi2_names, self.chi2_names)
                        unexpected = loaded_chi2_names.difference(
                            self.chi2_names).difference(self.derived_params)
                        if unexpected:
                            raise LoggedError(self.log,
                                              "Input samples have chi2 values "
                                              "that are not expected: %s ", unexpected)
                    else:
                        data_col_set = set(self.data.columns)
                        col_set = set(self.columns)
                        if data_col_set != col_set:
                            missing = set(self.columns).difference(self.data.columns)
                            unexpected = set(self.data.columns).difference(self.columns)
                            raise LoggedError(
                                self.log, (f"Bad column names! Missing {missing}; "
                                           f"unexpected: {unexpected}"))
                    self._n_last_out = len(self)
                except IOError:
                    if resuming:
                        self.log.info(
                            "Could not find a chain to resume. "
                            "Maybe burn-in didn't finish. Creating new chain file!")
                        resuming = False
                    elif load:
                        raise
            else:
                raise LoggedError(self.log,
                                  "No continuation possible if there is no output.")
        else:
            self._out_delete()
        if not resuming and not load:
            self.reset()
        # If loaded, check sample weights, consistent logp sums,
        # and temperature (ignores the given one)
        samples_loaded = len(self) > 0
        if samples_loaded:
            try:
                self.temperature = self._check_logps()
                if temperature is not None and \
                   not np.isclose(temperature, self.temperature):
                    raise LoggedError(
                        self.log,
                        "Sample temperature appears to be %r, but the collection was "
                        "explicitly initialized with temperature %r.",
                        self.temperature,
                        temperature,
                    )
                self._check_weights()
            except LoggedError as excpt:
                raise LoggedError(
                    self.log,
                    "Error when loading samples: %s",
                    str(excpt)
                ) from excpt
            self._drop_samples_null_weight()
            if self.is_tempered and not resuming:
                self.log.warning(
                    "The collection loaded has temperature != 1. "
                    "Keep that in mind when operating on it, or detemper (in-place) with "
                    "'SampleCollection.reset_temperature()'."
                )
        else:
            self.temperature = temperature if temperature is not None else 1
        # Prepare fast numpy cache
        self._icol = {col: i for i, col in enumerate(self.columns)}
        self._cache_reset()
        # Prepare txt formatter
        self.n_float = 8
        # Add to this 7 places: sign, leading 0's, exp with sign and 3 figures.
        width_col = lambda col: max(7 + self.n_float, len(col))
        self._numpy_fmts = ["%{}.{}".format(width_col(col), self.n_float) + "g"
                            for col in self.data.columns]
        self._header_formatter = [
            eval('lambda s, w=width_col(col): '  # pylint: disable=eval-used
                 '("{:>" + "{}".format(w) + "s}").format(s)',
                 {'width_col': width_col, 'col': col})
            for col in self.data.columns]

    def reset(self):
        """Create/reset the DataFrame."""
        self._cache_reset()
        self._data = pd.DataFrame(columns=self.columns, dtype=np.float64)
        if getattr(self, "file_name", None):
            self._n_last_out = 0

    def add(self, values: Union[Sequence[float], np.ndarray],
            logpost: Optional[Union[LogPosterior, float]] = None,
            logpriors: Optional[Sequence[float]] = None,
            loglikes: Optional[Sequence[float]] = None,
            derived: Optional[Sequence[float]] = None,
            weight: float = 1):
        """
        Adds a point to the collection.

        If `logpost` can be :class:`~model.LogPosterior`, float or None (in which case,
        `logpriors`, `loglikes` are both required).

        If the weight is not specified, it is assumed to be 1.
        """
        logposterior = self._check_before_adding(
            values, logpost=logpost, logpriors=logpriors,
            loglikes=loglikes, derived=derived, weight=weight)
        self._cache_add(values, logposterior=logposterior, weight=weight)

    def _check_before_adding(self, values: Union[Sequence[float], np.ndarray],
                             logpost: Optional[Union[LogPosterior, float]] = None,
                             logpriors: Optional[Sequence[float]] = None,
                             loglikes: Optional[Sequence[float]] = None,
                             derived: Optional[Sequence[float]] = None,
                             weight: float = 1
                             ) -> LogPosterior:
        """
        Checks that the arguments of collection.add are correctly formatted.

        Returns a :class:`~model.LogPosterior` dataclass (unchanged if one passed).
        """
        if weight is not None and weight <= 0:
            raise LoggedError(self.log, "Weights must be positive. Got %r", weight)
        if len(values) != len(self.sampled_params):
            raise LoggedError(
                self.log, "Got %d values for the sampled parameters. Should be %d.",
                len(values), len(self.sampled_params))
        if derived is not None:
            if len(derived) != len(self.derived_params):
                raise LoggedError(
                    self.log, "Got %d values for the derived parameters. Should be %d.",
                    len(derived), len(self.derived_params))
        if isinstance(logpost, LogPosterior):
            # If priors and likes passed, check consistency
            if logpriors is not None:
                if not np.allclose(logpriors, logpost.logpriors):
                    raise LoggedError(
                        self.log,
                        "logpriors not consistent with LogPosterior object passed.")
            if loglikes is not None:
                if not np.allclose(loglikes, logpost.loglikes):
                    raise LoggedError(
                        self.log,
                        "loglikes not consistent with LogPosterior object passed.")
            if derived is not None:
                # A simple np.allclose is not enough, because np.allclose([1], []) = True!
                if len(derived) != len(logpost.derived) or \
                        not np.allclose(derived, logpost.derived):
                    raise LoggedError(
                        self.log,
                        "derived params not consistent with those of LogPosterior object "
                        "passed.")
            return_logpost = logpost
        elif isinstance(logpost, float) or logpost is None:
            try:
                return_logpost = LogPosterior(
                    logpriors=logpriors,  # type: ignore
                    loglikes=loglikes,  # type: ignore
                    derived=derived,  # type: ignore
                )
            except ValueError as valerr:
                # missing logpriors/loglikes if logpost is None,
                # or inconsistent sum if logpost given
                raise LoggedError(self.log, str(valerr)) from valerr
        else:
            raise LoggedError(
                self.log, "logpost must be a LogPosterior object, a number or None (in "
                          "which case logpriors and loglikes are needed).")
        return return_logpost

    def _cache_reset(self):
        self._cache = np.full((self.cache_size, len(self.columns)), np.nan)
        self._cache_last = -1

    def _cache_add(self, values: Union[Sequence[float], np.ndarray],
                   logposterior: LogPosterior, weight: float = 1):
        """
        Adds the given point to the cache. Dumps and resets the cache if full.
        """
        if self._cache_last == self.cache_size - 1:
            self._cache_dump()
        self._cache_add_row(
            self._cache_last + 1, values, logposterior=logposterior, weight=weight)
        self._cache_last += 1

    def _cache_add_row(self, pos: int, values: Union[Sequence[float], np.ndarray],
                       logposterior: LogPosterior, weight: float = 1):
        """
        Adds the given point to the cache at the given position.
        """
        self._cache[pos, self._icol[OutPar.weight]] = weight if weight is not None else 1
        self._cache[pos, self._icol[OutPar.minuslogpost]] = \
            -apply_temperature(logposterior.logpost, self.temperature)
        for name, value in zip(self.sampled_params, values):
            self._cache[pos, self._icol[name]] = value
        if logposterior.logpriors is not None:
            for name, value in zip(self.minuslogprior_names, logposterior.logpriors):
                self._cache[pos, self._icol[name]] = -value
            self._cache[pos, self._icol[OutPar.minuslogprior]] = - logposterior.logprior
        if logposterior.loglikes is not None:
            for name, value in zip(self.chi2_names, logposterior.loglikes):
                self._cache[pos, self._icol[name]] = -2 * value
            self._cache[pos, self._icol[OutPar.chi2]] = -2 * logposterior.loglike
        if len(logposterior.derived):
            for name, value in zip(self.derived_params, logposterior.derived):
                self._cache[pos, self._icol[name]] = value

    def _cache_dump(self):
        """
        Dumps the cache into the pandas table (unless empty).
        """
        if self._cache_last == -1:
            return
        self._enlarge(self._cache_last + 1)
        self._data.iloc[len(self._data) - (self._cache_last + 1):len(self._data)] = \
            self._cache[:self._cache_last + 1]
        self._cache_reset()

    def _check_logps(self, temperature_only=False):
        """
        Checks the correct sums for the logpriors and likelihoods' chi-squared's, as well
        as the posterior temperature, which is returned.

        Raises ``LoggedError`` if the sums or the temperature of the sample are not
        consistent.
        """
        try:
            temperature = compute_temperature(
                -self["minuslogpost"], -self["minuslogprior"], -self["chi2"] * 0.5,
                check=True)
        except AssertionError as excpt:
            raise LoggedError(
                self.log, "The sample seems to have an inconsistent temperature.") \
                from excpt
        if not temperature_only:
            if not np.allclose(
                    self[OutPar.minuslogprior],
                    np.sum(
                        np.atleast_2d(
                            self[self.minuslogprior_names].to_numpy(dtype=np.float64)),
                        axis=-1,
                    )
            ):
                raise LoggedError(
                    self.log,
                    "The sum of logpriors in the sample is not consistent."
                )
            if not np.allclose(
                    self[OutPar.chi2],
                    np.sum(
                        np.atleast_2d(self[self.chi2_names].to_numpy(dtype=np.float64)),
                        axis=-1,
                    )
            ):
                raise LoggedError(
                    self.log,
                    "The sum of likelihood's chi2's in the sample is not consistent."
                )
        if np.isclose(temperature, 1):
            temperature = 1
        return temperature

    def _check_weights(
            self,
            weights: Optional[np.ndarray] = None,
            length: Optional[int] = None
    ):
        """
        Checks correct length, shape and signs of the ``weights``.

        If no weights passed, checks internal consistency.

        If ``length`` passed, checks for specific length of the weights vector.

        Raises ``LoggedError`` if the weights are badly arranged or invalid.
        """
        if weights is None:
            weights_array = self[OutPar.weight].to_numpy(dtype=np.float64)
        else:
            weights_array = np.atleast_1d(weights)
            if len(weights_array.shape) != 1:
                raise LoggedError(
                    self.log,
                    "The shape of the weights is wrong. Expected a 1d array, "
                    "but got shape %r.",
                    weights_array.shape
                )
            check_length = len(self) if length is None else length
            if len(weights_array) != check_length:
                raise LoggedError(
                    self.log,
                    "The length of the weights vector is wrong. Expected %d but got %d.",
                    check_length,
                    len(weights_array)
                )
        if np.any(weights_array < 0):
            raise LoggedError(
                self.log,
                "The weight vector contains negative elements."
            )

    @property
    def is_tempered(self) -> bool:
        """
        Whether the sample was obtained by drawing from a different-temperature
        distribution.
        """
        return self.temperature != 1

    @property
    def has_int_weights(self) -> bool:
        """
        Whether weights are integer.
        """
        weights = self[OutPar.weight]
        return np.allclose(np.round(weights), weights)

    def _detempered_weights(self):
        """Computes the detempered weights."""
        if self.temperature == 1:
            return self._data[OutPar.weight].to_numpy(dtype=np.float64)
        return (
            self._data[OutPar.weight].to_numpy(dtype=np.float64) *
            detempering_weights_factor(
                -self._data[OutPar.minuslogpost].to_numpy(dtype=np.float64),
                self.temperature
            )
        )

    def _detempered_minuslogpost(self):
        """Computes the detempered -log-posterior."""
        if self.temperature == 1:
            return self._data[OutPar.minuslogpost].to_numpy(dtype=np.float64)
        return -remove_temperature(
            -self._data[OutPar.minuslogpost].to_numpy(dtype=np.float64),
            self.temperature
        )

    def reset_temperature(self):
        """
        Drops the information about sampling temperature: ``weight`` and ``minuslogpost``
        columns will now correspond to those of a unit-temperature posterior sample.

        This cannot be undone: (e.g. recovering original integer tempered weights).
        You may want to call this method on a copy (see :func:`SampleCollection.copy`).
        """
        self._cache_dump()
        if self.temperature == 1:
            return
        self._data[OutPar.weight] = self._detempered_weights()
        self._drop_samples_null_weight()
        self._data[OutPar.minuslogpost] = self._detempered_minuslogpost()
        self.temperature = 1

    def _enlarge(self, n):
        """
        Enlarges the DataFrame by `n` rows.
        """
        self._data = pd.concat([
            self._data, pd.DataFrame(
                np.nan, columns=self._data.columns,
                index=np.arange(len(self._data), len(self._data) + n))])

    def _append(self, collection):
        """
        Append another collection.
        Internal method: does not check for consistency!
        """
        self._data = pd.concat([self.data, collection.data],
                               ignore_index=True)

    def __len__(self):
        return len(self._data) + (self._cache_last + 1)

    def __bool__(self):
        return len(self) != 0

    @property
    def n_last_out(self):
        """Index of the last point saved to the output."""
        return self._n_last_out

    @property  # type: ignore
    @ensure_cache_dumped
    def data(self):
        """Pandas' ``DataFrame`` containing the sample collection."""
        return self._data

    # Make the dataframe printable (but only the filled ones!)
    def __repr__(self):
        return self.data.__repr__()

    # Make the dataframe iterable over rows
    def __iter__(self):
        return self.data.iterrows()

    # Accessing the dataframe
    def __getitem__(self, *args):
        """
        Direct access to the DataFrame, ensuring cache has been dumped.

        Returns views or copies as Pandas would do.
        """
        return self.data.__getitem__(*args)

    # MARKED FOR DEPRECATION IN v3.2
    @property
    def values(self) -> np.ndarray:
        """
        Returns the sample collection as a numpy array; to be deprecated in favour of
        ``Collection.to_numpy``, following Pandas.
        """
        print("*WARNING* Following Pandas, the 'Collection.values()' method will soon be "
              "deprecated in favour of 'Collection.to_numpy()'.")
        # BEHAVIOUR TO BE REPLACED BY AN ERROR
        return self.data.to_numpy(dtype=np.float64)
        # END OF DEPRECATION BLOCK

    def to_numpy(self, dtype=None, copy=False) -> np.ndarray:
        """Returns the sample collection as a numpy array."""
        return self.data.to_numpy(copy=copy, dtype=dtype or np.float64)

    def _copy(self, data=None, empty=False) -> 'SampleCollection':
        """
        Returns a copy of the collection.

        If ``empty=True`` (default ``False``), returns an empty copy.

        If data specified (default: None), the copy returned contains the given data;
        no checks are performed on given data, so use with care (e.g. use with a slice of
        ``self.data``).
        """
        current_data = self.data
        if data is None:
            data = current_data
        # Avoids creating an unnecessary copy of the data, to save memory
        delattr(self, "_data")
        self_copy = deepcopy(self)  # deletes logger (see HasLogger.__deepcopy___)
        self_copy.set_logger()
        setattr(self, "_data", current_data)
        if empty:
            self_copy.reset()
        else:
            setattr(self_copy, "_data", data.copy())
            setattr(self_copy, "_n", data.last_valid_index() + 1)
        return self_copy

    # Dummy function to avoid exposing `data` kwarg, since no checks are performed on it.
    def copy(self, empty=False) -> 'SampleCollection':
        """
        Returns a copy of the collection.

        If ``empty=True`` (default ``False``), returns an empty copy.
        """
        return self._copy(empty=empty)

    def _weights_for_stats(
            self,
            first: Optional[int] = None,
            last: Optional[int] = None,
            weights: Optional[np.ndarray] = None,
            tempered: bool = False,
    ) -> Tuple[np.ndarray, bool]:
        """
        Returns ``(weights, are_int)``, where weights can be used for computation of
        statistical quantities such as mean and covariance, and ``are_int`` is ``True``
        if the returned weights are known to be integer, and ``False`` if it cannot be
        guaranteed.

        If ``tempered=True`` (default ``False``) returns the weights of the tempered
        posterior ``p**(1/temperature)``.

        Custom weights can be passed with the argument ``weights``. In that case, internal
        weights are ignored, and some checks are performed on the passed ones.
        """
        if weights is not None:
            first = first or 0
            last = last or len(self)
            self._check_weights(weights, length=last - first)
            weights /= max(weights)
            return weights, np.allclose(np.round(weights), weights)
        if self.is_tempered and not tempered:
            # For sure the weights are not integer
            return self._detempered_weights()[first:last], False
        return (
            self[OutPar.weight][first:last].to_numpy(dtype=np.float64),
            self.has_int_weights
        )

    def mean(
            self,
            first: Optional[int] = None,
            last: Optional[int] = None,
            weights: Optional[np.ndarray] = None,
            derived: bool = False,
            tempered: bool = False
    ) -> np.ndarray:
        """
        Returns the (weighted) mean of the parameters in the chain,
        between `first` (default 0) and `last` (default last obtained),
        optionally including derived parameters if `derived=True` (default `False`).

        Custom weights can be passed with the argument ``weights``.

        If ``derived`` is ``True`` (default ``False``), the means of the derived
        parameters are included in the returned vector.

        If ``tempered=True`` (default ``False``) returns the mean of the tempered
        posterior ``p**(1/temperature)``.

        NB: For tempered samples, if passed ``tempered=False`` (default), detempered
        weights are computed on-the-fly. If this or any other function returning
        untempered statistical quantities of a tempered sample is expected to be called
        repeatedly, it would be more efficient to detemper the collection first with
        :func:`SampleCollection.reset_temperature`, and call these methods on the returned
        Collection.
        """
        if not self:
            raise LoggedError(self.log, "Collection is empty. Cannot compute mean.")
        weights_mean, _ = self._weights_for_stats(
            first, last, weights=weights, tempered=tempered)
        return np.average(self[list(self.sampled_params) +
                               (list(self.derived_params) if derived else [])]
                          [first:last].to_numpy(dtype=np.float64).T, weights=weights_mean,
                          axis=-1
                          )

    def cov(
            self,
            first: Optional[int] = None,
            last: Optional[int] = None,
            weights: Optional[np.ndarray] = None,
            derived: bool = False,
            tempered: bool = False
    ) -> np.ndarray:
        """
        Returns the (weighted) covariance matrix of the parameters in the chain,
        between `first` (default 0) and `last` (default last obtained),
        optionally including derived parameters if `derived=True` (default `False`).

        Custom weights can be passed with the argument ``weights``.

        If ``derived`` is ``True`` (default ``False``), the covariances of/with the
        derived parameters are included in the returned matrix.

        If ``tempered=True`` (default ``False``) returns the covariances of the tempered
        posterior ``p**(1/temperature)``.

        NB: For tempered samples, if passed ``tempered=False`` (default), detempered
        weights are computed on-the-fly. If this or any other function returning
        untempered statistical quantities of a tempered sample is expected to be called
        repeatedly, it would be more efficient to detemper the collection first with
        :func:`SampleCollection.reset_temperature`, and call these methods on the returned
        Collection.
        """
        if not self:
            raise LoggedError(self.log, "Collection is empty. Cannot compute cov.")
        weights_cov, are_int = self._weights_for_stats(
            first, last, weights=weights, tempered=tempered)
        weight_type_kwarg = "fweights" if are_int else "aweights"
        return np.atleast_2d(np.cov(  # type: ignore
            self[list(self.sampled_params) +
                 (list(self.derived_params) if derived else [])][first:last].to_numpy(
                dtype=np.float64).T,
            **{weight_type_kwarg: weights_cov}))

    def _drop_samples_null_weight(self):
        """Removes from the DataFrame all samples that have 0 weight."""
        self._data = self.data[self._data.weight > 0].reset_index(drop=True)
        self._n = self._data.last_valid_index() + 1

    def reweight(self, importance_weights, check=True):
        """
        Reweights the sample with the given ``importance_weights``.

        Temperature information is dropped.

        This cannot be fully undone (e.g. recovering original integer weights).
        You may want to call this method on a copy (see :func:`SampleCollection.copy`).

        For the sake of speed, length and positivity checks on the importance weights can
        be skipped with ``check=False`` (default ``True``).
        """
        self.reset_temperature()  # includes a self._cache_dump()
        if check:
            self._check_weights(importance_weights)
        self._data[OutPar.weight] *= importance_weights
        self._drop_samples_null_weight()

    def filtered_copy(self, where) -> 'SampleCollection':
        """Returns a copy of the collection with some condition ``where`` imposed."""
        return self._copy(self.data[where].reset_index(drop=True))

    def skip_samples(self, skip: float, inplace: bool = False) -> 'SampleCollection':
        """
        Skips some initial samples, or an initial fraction of them.

        For collections coming from a Nested Sampler, prints a warning and does nothing.

        Parameters
        ----------
        skip: float
            Specified the amount of initial samples to be skipped, either directly if
            ``skip>1`` (rounded up to next integer), or as a fraction if ``0<skip<1``.

        inplace: bool, default: False
            If True, returns a copy of the collection.

        Returns
        -------
        SampleCollection
            The original collection with skipped initial samples (``inplace=True``) or
            a copy of it (``inplace=False``).

        Raises
        ------
        LoggedError
            If badly defined ``skip`` value.
        """
        if skip == 0:
            return self if inplace else self.copy()
        if self.sample_type == "nested":
            self.log.warning(
                "Cannot skip initial samples from Nested Sampling samples. Doing nothing"
            )
            return self if inplace else self.copy()
        if not isinstance(skip, Number) or skip < 0:
            raise LoggedError(
                self.log,
                "Number of fraction of skipped initial samples must be positive. Got %r",
                skip
            )
        if 0 < skip < 1:
            skip = int(np.round(skip * len(self)))
        skip = int(np.round(skip))
        if inplace:
            self._data = self.data[skip:].reset_index(drop=True)
            self._n = self._data.last_valid_index() + 1
            return self
        else:
            return self.filtered_copy(slice(skip, None))

    def thin_samples(self, thin: int, inplace: bool = False) -> 'SampleCollection':
        """
        Thins the sample collection by some factor ``thin>1``.

        Parameters
        ----------
        thin: int
            Thin factor, must be ``>1``.
        inplace: bool, default: False
            If True, returns a copy of the collection.

        Returns
        -------
        SampleCollection
            Thinned version of the original collection (``inplace=True``) or a copy of it
            (``inplace=False``).

        Raises
        ------
        LoggedError
            If badly defined ``thin`` value.
        """
        if thin == 1:
            return self if inplace else self.copy()
        if thin != int(thin) or thin < 1:
            raise LoggedError(self.log, "Thin factor must be a positive integer, got %s",
                              thin)
        thin = int(thin)
        try:
            if hasattr(WeightedSamples, "thin_indices_and_weights"):
                unique, counts = \
                    WeightedSamples.thin_indices_and_weights(
                        thin, self[OutPar.weight].to_numpy(dtype=np.float64))
            else:
                raise LoggedError(self.log, "Thinning requires GetDist 1.2+", )
        except WeightedSampleError as e:
            raise LoggedError(self.log, "Error thinning: %s", e) from e
        else:
            data = self._data.iloc[unique, :].copy()
            # Produces pandas warning (pandas<2.0). Safe to ignore. Delete later.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=FutureWarning)
                data.iloc[:, 0] = counts
            data.reset_index(drop=True, inplace=True)
            if inplace:
                self._data = data
                self._n = self._data.last_valid_index() + 1
            else:
                return self._copy(data)
        return self

    def bestfit(self):
        """Best fit (maximum likelihood) sample. Returns a copy."""
        return self.data.loc[self.data[OutPar.chi2].astype(np.float64).idxmin()].copy()

    def MAP(self):
        """Maximum-a-posteriori (MAP) sample. Returns a copy."""
        return self.data.loc[self.data[OutPar.minuslogpost].astype(np.float64).idxmin()]\
                        .copy()

    def _sampled_to_getdist(
            self,
            first: Optional[int] = None,
            last: Optional[int] = None,
            tempered: bool = False
    ) -> MCSamples:
        """
        Barebones interface with getdist. Internal use only!
        Use :func:`SampleCollection.to_getdist` instead.
        """
        names = list(self.sampled_params)
        weights, _ = self._weights_for_stats(first, last, tempered=tempered)
        if self.is_tempered and not tempered:
            minuslogposts = self._detempered_minuslogpost()[first:last]
        else:
            minuslogposts = self.data[OutPar.minuslogpost].to_numpy(
                dtype=np.float64)[first:last]
        # No logging of warnings temporarily, so getdist won't complain unnecessarily
        with NoLogging():
            mcsamples = MCSamples(
                samples=self.data[names].to_numpy(dtype=np.float64)[first:last],
                weights=weights, loglikes=minuslogposts, names=names,
            )
        return mcsamples

    def to_getdist(
            self,
            label: Optional[str] = None,
            model: Optional[Model] = None,
    ) -> MCSamples:
        """
        Parameters
        ----------
        label: str, optional
            Legend label in ``GetDist`` plots (``name_tag`` in ``GetDist`` parlance).
        model: :class:`cobaya.model.Model`, optional
            `Model` with which the sample was created. Needed only if parameter labels or
            aliases have changed since the collection was generated.

        Returns
        -------
        :class:`getdist.MCSamples`
            This collection's equivalent :class:`getdist.MCSamples` object.

        Raises
        ------
        LoggedError
            Errors when processing the arguments.
        """
        if isinstance(model, Model):
            self._cache_aux_model_quantities(model)
        elif model is not None:
            LoggedError("Optional argument `model` must be a Cobaya Model instance.")
        used_names_dict = {p: p + ("*" if p not in self.sampled_params else "")
                           for p in self.data.columns[2:]}
        return MCSamples(
            samples=self[self.data.columns[2:]].to_numpy(np.float64, copy=True),
            weights=self[OutPar.weight].to_numpy(np.float64, copy=True),
            loglikes=self[OutPar.minuslogpost].to_numpy(np.float64, copy=True),
            temperature=self.temperature,
            sampler=deepcopy(self.sample_type),
            names=list(used_names_dict.values()),
            labels=[deepcopy(self._cached_labels[p]) for p in used_names_dict
                    if p in self._cached_labels],
            ranges=deepcopy(self._cached_ranges),
            renames=deepcopy(self._cached_renames),
            name_tag=label,
            label=deepcopy(self.name),
            # ini=ini,
            # settings=settings
        )

    # Saving and updating
    def _get_driver(self, method):
        return getattr(self, method + derived_par_name_separator + self.driver)

    # Load a pre-existing file
    def _out_load(self, **kwargs):
        self._get_driver("_load")(**kwargs)
        self._cache_reset()

    # Dump/update/delete collection

    def out_update(self):
        """Update the output file to the current state of the Collection."""
        self._get_driver("_update")()

    def _out_delete(self):
        self._get_driver("_delete")()

    # txt driver
    def _load__txt(self, skip=0):
        self.log.debug("Skipping %d rows", skip)
        if bool(skip) and self.sample_type == "nested":
            raise LoggedError(
                self.log,
                "Cannot skip samples from a sample of a nested sampler. "
                "Would lead to a non-fair sample."
            )
        self._data = load_DataFrame(self.file_name, skip=skip,
                                    root_file_name=self.root_file_name)
        self.log.info("Loaded %d sample points from '%s'", len(self._data),
                      self.file_name)

    def _dump__txt(self):
        self._dump_slice__txt(0, len(self))

    def _update__txt(self):
        self._dump_slice__txt(self.n_last_out, len(self))

    def _dump_slice__txt(self, n_min=None, n_max=None):
        if n_min is None or n_max is None:
            raise LoggedError(self.log, "Needs to specify the limit n's to dump.")
        if self._n_last_out == n_max:
            return
        self._n_last_out = n_max
        do_header = not n_min
        if do_header:
            if os.path.exists(self.file_name):
                raise LoggedError(self.log,
                                  "Output file %s already exists (report bug)",
                                  self.file_name)
            with open(self.file_name, "a", encoding="utf-8") as out:
                out.write("#" + " ".join(
                    f(col) for f, col
                    in zip(self._header_formatter, self.data.columns))[1:] + "\n")
        with open(self.file_name, "a", encoding="utf-8") as out:
            np.savetxt(out, self.data[n_min:n_max].to_numpy(dtype=np.float64),
                       fmt=self._numpy_fmts)

    def _delete__txt(self):
        try:
            os.remove(self.file_name)
        except OSError:
            pass

    # dummy driver
    def _dump__dummy(self):
        pass

    def _update__dummy(self):
        pass

    def _delete__dummy(self):
        pass

    # Make it picklable -- formatters are deleted
    # (they will be generated next time txt is dumped)
    def __getstate__(self):
        attributes = super().__getstate__().copy()
        for attr in ['_numpy_fmts', '_header_formatter']:
            try:
                del attributes[attr]
            except KeyError:
                pass
        return attributes


class OneSamplePoint:
    """
    Wrapper to hold a single point, e.g. the current point of an MCMC.
    Alternative to :class:`~collection.OnePoint`, faster but with less functionality.

    For tempered samples, stores the weight and -logp of the tempered posterior (but
    untempered priors and likelihoods).
    """

    results: LogPosterior
    values: np.ndarray
    weight: int

    def __init__(self, model, temperature=1, output_thin=1):
        self.sampled_params = list(model.parameterization.sampled_params())
        self.temperature = temperature
        self.output_thin = output_thin
        self._added_weight = 0

    def add(self, values, logpost: LogPosterior):
        """Add/override the sampled point."""
        self.values = values
        if not isinstance(logpost, LogPosterior):
            raise ValueError("`logpost` argument must be LogPosterior instance.")
        self.results = logpost
        self.weight = 1

    @property
    def logpost(self):
        """:class:`~model.Logposterior` instance of the sampled point."""
        return self.results.logpost

    def add_to_collection(self, collection: SampleCollection):
        """
        Adds this point at the end of a given collection.

        It is assumed that both this instance and the collection passed were
        initialised with the same :class:`model.Model` (no checks are performed).
        """
        if self.output_thin > 1:
            self._added_weight += self.weight
            if self._added_weight >= self.output_thin:
                weight = self._added_weight // self.output_thin
                self._added_weight %= self.output_thin
            else:
                return False
        else:
            weight = self.weight
        collection.add(self.values, logpost=self.results, weight=weight)
        return True

    def __str__(self):
        return ", ".join(
            ['%s:%.7g' % (k, v) for k, v in zip(self.sampled_params, self.values)])


class OnePoint(SampleCollection):
    """
    Wrapper of :class:`~collection.SampleCollection` to hold a single point,
    e.g. the best-fit point of a minimization run (not used by default MCMC).
    """

    def __getitem__(self, columns, *args):
        if isinstance(columns, str):
            return self.data.values[0, self.data.columns.get_loc(columns)]
        try:
            return self.data.values[0,
                                    [self.data.columns.get_loc(c) for c in columns]]
        except KeyError as excpt:
            raise ValueError("Some of the indices are not valid columns.") from excpt

    def add(self, *args, **kwargs):
        """Add/override the sampled point."""
        self.reset()  # resets the DataFrame, so never goes beyond 1 point
        super().add(*args, **kwargs)

    def increase_weight(self, increase=1):
        """Increase the weight of the point by ``increase`` (default: 1)."""
        # For some reason, faster than `self.data[Par.weight] += increase`
        self.data.at[0, OutPar.weight] += increase

    # Restore original __repr__ (here, there is only 1 sample)
    def __repr__(self):
        return self.data.__repr__()

    # Dumper changed: always force to print the single element.
    def _update__txt(self):
        self._dump_slice__txt(0, 1)



from copy import deepcopy
from .test_cosmo_planck_2015 import params_lowTEB_highTTTEEE
from .common_cosmo import body_of_test


# Pantheon (alpha and beta not used - no nuisance parameters), fast
def test_sn_pantheon_camb(packages_path, skip_not_installed):
    lik = "sn.pantheon"
    info_likelihood = {lik: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2_sn_pantheon,
                 skip_not_installed=skip_not_installed)


def test_sn_pantheon_classy(packages_path, skip_not_installed):
    lik = "sn.pantheon"
    info_likelihood = {lik: {}}
    info_theory = {"classy": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2_sn_pantheon,
                 skip_not_installed=skip_not_installed)


# JLA
def test_sn_jla_camb(packages_path, skip_not_installed):
    best_fit_test = deepcopy(params_lowTEB_highTTTEEE)
    best_fit_test.update(best_fit_sn)
    lik = "sn.jla"
    info_likelihood = {lik: {}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit_test, info_likelihood, info_theory, chi2_sn_jla,
                 skip_not_installed=skip_not_installed)


def test_sn_jla_classy(packages_path, skip_not_installed):
    best_fit_test = deepcopy(params_lowTEB_highTTTEEE)
    best_fit_test.update(best_fit_sn)
    lik = "sn.jla"
    info_likelihood = {lik: {}}
    info_theory = {"classy": None}
    body_of_test(packages_path, best_fit_test, info_likelihood, info_theory, chi2_sn_jla,
                 skip_not_installed=skip_not_installed)


# JLA marginalized over alpha, beta
def test_sn_jla_lite_camb(packages_path, skip_not_installed):
    lik = "sn.jla_lite"
    info_likelihood = {lik: {"marginalize": True}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2_sn_jla_lite,
                 skip_not_installed=skip_not_installed)


# JLA marginalized over alpha, beta (slow version!)
def test_sn_jla_lite_slow_camb(packages_path, skip_not_installed):
    lik = "sn.jla_lite"
    info_likelihood = {lik: {"marginalize": True, "precompute_covmats": False}}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit, info_likelihood, info_theory, chi2_sn_jla_lite,
                 skip_not_installed=skip_not_installed)


def test_sn_pantheon_Mb(packages_path, skip_not_installed):
    best_fit_test = deepcopy(params_lowTEB_highTTTEEE)
    best_fit_test.update(best_fit_Mb)
    info_likelihood = {"sn.pantheon": {"use_abs_mag": True}, "H0.riess2020Mb": None}
    info_theory = {"camb": None}
    body_of_test(packages_path, best_fit_test, info_likelihood, info_theory,
                 chi2_sn_pantheon_Mb, skip_not_installed=skip_not_installed)


# BEST FIT AND REFERENCE VALUES ##########################################################

best_fit = deepcopy(params_lowTEB_highTTTEEE)
best_fit_sn = {"alpha_jla": 0.1325237, "beta_jla": 2.959805}
best_fit_Mb = {"Mb": -19.2}

chi2_sn_pantheon = {"sn.pantheon": 1035.30, "tolerance": 0.1}
chi2_sn_pantheon_Mb = {"sn.pantheon": 4025.30, "H0.riess2020Mb": 1.65, "tolerance": 0.1}
chi2_sn_jla = {"sn.jla": 700.582, "tolerance": 0.1}
chi2_sn_jla_lite = {"sn.jla_lite": 706.882, "tolerance": 0.1}



r"""
.. module:: one

:Synopsis: Unit likelihood: always outputs :math:`0 = \log(1)`.
:Author: Jesus Torrado

"""

# Global
from random import random
from typing import Optional

# Local
from cobaya.likelihood import AbsorbUnusedParamsLikelihood


class one(AbsorbUnusedParamsLikelihood):
    """
    Likelihood that evaluates to 1.
    """
    noise: Optional[float]

    def initialize(self):
        if self.noise:
            self.logp = self.logp_noise
        else:
            self.logp = self.logp_one

    def logp_one(self, **_params_values):
        self.wait()
        return 0.

    def logp_noise(self, **_params_values):
        self.wait()
        if self.noise:
            return self.noise * random()
        else:
            return 0



"""
.. module:: doc

:Synopsis: Show defaults options for components.
:Author: Jesus Torrado

"""

# Global
from pprint import pformat

# Local
from cobaya.tools import warn_deprecation, get_available_internal_class_names, \
    similar_internal_class_names
from cobaya.component import get_component_class, ComponentNotFoundError
from cobaya.conventions import subfolders, kinds
from cobaya.input import get_default_info
from cobaya.log import logger_setup, get_logger

_indent = 2 * " "


# Command-line script ####################################################################

def doc_script(args=None):
    """Command line script for the documentation."""
    warn_deprecation()
    logger_setup()
    logger = get_logger("doc")
    # Parse arguments
    import argparse
    parser = argparse.ArgumentParser(
        prog="cobaya doc", description="Prints defaults for Cobaya's components.")
    parser.add_argument("component", action="store", nargs="?", default="",
                        metavar="component_name",
                        help=("The component whose defaults are requested. "
                              "Pass a component kind (sampler, theory, likelihood) to "
                              "list all available (internal) ones, pass nothing to list "
                              "all available (internal) components of all kinds."))
    parser.add_argument("-p", "--python", action="store_true", default=False,
                        help="Request Python instead of YAML.")
    expand_flag, expand_flag_ishort = "expand", 1
    parser.add_argument("-" + expand_flag[expand_flag_ishort], "--" + expand_flag,
                        action="store_true", default=False, help="Expand YAML defaults.")
    arguments = parser.parse_args(args)
    # Nothing passed: list all
    if not arguments.component:
        msg = "Available components: (some may need external code/data)"
        print(msg + "\n" + "-" * len(msg))
        for kind in kinds:
            print("%s:" % kind)
            print(_indent + ("\n" + _indent).join(
                get_available_internal_class_names(kind)))
        return
    # A kind passed (plural or singular): list all of that kind
    if arguments.component.lower() in subfolders.values():
        arguments.component = next(
            k for k in subfolders if arguments.component == subfolders[k])
    if arguments.component.lower() in kinds:
        print("%s:" % arguments.component.lower())
        print(_indent +
              ("\n" + _indent).join(
                  get_available_internal_class_names(arguments.component.lower())))
        return
    # Otherwise, try to identify the component
    try:
        cls = get_component_class(arguments.component, logger=logger)
    except ComponentNotFoundError:
        suggestions = similar_internal_class_names(arguments.component)
        logger.error(
            f"Could not identify component '{arguments.component}'. "
            f"Did you mean any of the following? {suggestions} (mind capitalization!)")
        return 1
    to_print = get_default_info(
        cls, return_yaml=not arguments.python, yaml_expand_defaults=arguments.expand)
    if arguments.python:
        print(pformat({cls.get_kind(): {arguments.component: to_print}}))
    else:
        print(cls.get_kind() + ":\n" + _indent + arguments.component + ":\n" +
              2 * _indent + ("\n" + 2 * _indent).join(to_print.split("\n")))
        if "!defaults" in to_print:
            print("# This file contains defaults. "
                  "To populate them, use the flag --%s (or -%s)." % (
                      expand_flag, expand_flag[expand_flag_ishort]))


if __name__ == '__main__':
    doc_script()



try:
    from .gui import gui_script
except ImportError:
    # PySide not installed, but pass for now (will fail at GUI initialization)
    pass
from .autoselect_covmat import get_best_covmat, get_best_covmat_ext
from .create_input import create_input
from .input_database import planck_base_model, base_precision, planck_precision, \
    planck_lss_precision, cmb_precision
from .input_database import install_basic, install_tests, _combo_dict_text



from cobaya.likelihoods.base_classes import Planck2015Clik


class planck_2015_lowTEB(Planck2015Clik):
    r"""
    Low-$\ell$ polarization likelihood of Planck's 2015 data release
    \cite{Aghanim:2015xee}.
    """
    pass



"""
Automatic tests of the cosmo_model example in the documentation,
to make sure it remains up to date.
"""

import os
import numpy as np
from io import StringIO

from .common import process_packages_path, stdout_redirector
from .conftest import install_test_wrapper

tests_folder = os.path.dirname(os.path.realpath(__file__))
docs_folder = os.path.join(tests_folder, "..", "docs")
docs_src_folder = os.path.join(docs_folder, "src_examples", "cosmo_model")
docs_img_folder = os.path.join(docs_folder, "img")

# Number of possible different pixels
pixel_tolerance = 0.995


def test_cosmo_docs_model_classy(packages_path, skip_not_installed):
    packages_path = process_packages_path(packages_path)
    # Go to the folder containing the python code
    cwd = os.getcwd()
    try:
        os.chdir(docs_src_folder)
        globals_example = {}
        exec(open(os.path.join(docs_src_folder, "1.py")).read(), globals_example)
        globals_example["info"]["packages_path"] = packages_path
        install_test_wrapper(
            skip_not_installed, exec,
            open(os.path.join(docs_src_folder, "2.py")).read(), globals_example)
        stream = StringIO()
        with stdout_redirector(stream):
            exec(open(os.path.join(docs_src_folder, "3.py")).read(), globals_example)
        # Comparing text output for this cell -- only derived parameter values
        out_filename = "3.out"
        derived_line_old, derived_line_new = map(
            lambda lines: next(line for line in lines[::-1] if line),
            [open(os.path.join(docs_src_folder, out_filename)).readlines(),
             stream.getvalue().split("\n")])
        derived_params_old, derived_params_new = map(
            lambda x: eval(x[x.find("{"):]), [derived_line_old, derived_line_new])
        oldvals = list(derived_params_old.values())
        newvals = [derived_params_new[v] for v in derived_params_old]
        assert np.allclose(oldvals, newvals), (
                "Wrong derived parameters line:\nBEFORE: %s\nNOW:    %s" %
                (derived_line_old, derived_line_new))
        # Not testing figures for now (not robust)
        # Instead, let's just check that no error is raised
        for filename in ["4.py", "5.py"]:
            try:
                exec(open(os.path.join(docs_src_folder, filename)).read(),
                     globals_example)
            except:
                assert False, "File %s failed." % filename
        #        if test_figs:
        #            # Compare plots
        #            pre = "cosmo_model_"
        #            for filename, imgname in zip(["4.py", "5.py"], ["cltt.png", "omegacdm.png"]):
        #                exec(open(os.path.join(docs_src_folder, filename)).read(), globals_example)
        #                old_img = imread(os.path.join(docs_img_folder, pre + imgname)).astype(float)
        #                new_img = imread(imgname).astype(float)
        #                npixels = (lambda x: x.shape[0] + x.shape[1])(old_img)
        #                assert np.count_nonzero(old_img == new_img) / (4 * npixels) >= pixel_tolerance, (
        #                        "Images '%s' are too different!" % imgname)
        # Passing the model to a sampler
    finally:
        # Back to the working directory of the tests, just in case
        os.chdir(cwd)



"""
.. module:: post

:Synopsis: Post-processing functions
:Author: Jesus Torrado and Antony Lewis

"""

import os
import time
from itertools import chain
from typing import List, Union, Optional, Tuple, TypedDict
import numpy as np

from cobaya import mpi
from cobaya.collection import SampleCollection
from cobaya.conventions import prior_1d_name, OutPar, get_chi2_name, \
    undo_chi2_name, get_minuslogpior_name, separator_files, minuslogprior_names, \
    packages_path_input
from cobaya.input import update_info, add_aggregated_chi2_params, load_input_dict
from cobaya.log import logger_setup, get_logger, is_debug, LoggedError
from cobaya.model import Model, DummyModel
from cobaya.output import get_output
from cobaya.parameterization import Parameterization
from cobaya.parameterization import is_fixed_or_function_param, is_sampled_param, \
    is_derived_param
from cobaya.prior import Prior
from cobaya.tools import progress_bar, recursive_update, deepcopy_where_possible, \
    str_to_list
from cobaya.typing import ExpandedParamsDict, ModelBlock, ParamValuesDict, InputDict, \
    PostDict


class PostResultDict(TypedDict):
    sample: Union[SampleCollection, List[SampleCollection]]
    stats: ParamValuesDict
    logpost_weight_offset: float
    weights: Union[np.ndarray, List[np.ndarray]]


_minuslogprior_1d_name = get_minuslogpior_name(prior_1d_name)


class OutputOptions:
    default_post_cache_size = 2000
    # to reweight we need to know absolute size of log likelihoods
    # but want to dump out regularly, so set _reweight_after as minimum to check first
    reweight_after = 100
    output_inteveral_s = 60


def value_or_list(lst: list):
    if len(lst) == 1:
        return lst[0]
    else:
        return lst


@mpi.sync_state
def post(info_or_yaml_or_file: Union[InputDict, str, os.PathLike],
         sample: Union[SampleCollection, List[SampleCollection], None] = None
         ) -> Tuple[InputDict, PostResultDict]:
    info = load_input_dict(info_or_yaml_or_file)
    # MARKED FOR DEPRECATION IN v3.2
    if info.get("debug_file"):
        raise LoggedError("'debug_file' has been deprecated. If you want to "
                          "save the debug output to a file, use 'debug: [filename]'.")
    # END OF DEPRECATION BLOCK
    logger_setup(info.get("debug"))
    log = get_logger(__name__)
    info_post: PostDict = info.get("post") or {}
    if not info_post:
        raise LoggedError(log, "No 'post' block given. Nothing to do!")
    if mpi.is_main_process() and info.get("resume"):
        log.warning("Resuming not implemented for post-processing. Re-starting.")
    if not info.get("output") and info_post.get("output") \
            and not info.get("params"):
        raise LoggedError(log, "The input dictionary must be a full option "
                               "dictionary, or have an existing 'output' root to load "
                               "previous settings from ('output' to read from is in the "
                               "main block not under 'post'). ")
    # 1. Load existing sample
    output_in = get_output(prefix=info.get("output"))
    if output_in:
        info_in = output_in.load_updated_info() or update_info(info)
    else:
        info_in = update_info(info)
    params_in: ExpandedParamsDict = info_in["params"]  # type: ignore
    dummy_model_in = DummyModel(params_in, info_in.get("likelihood", {}),
                                info_in.get("prior"))

    in_collections = []
    thin = info_post.get("thin", 1)
    skip = info_post.get("skip", 0)
    if info.get('thin') is not None or info.get('skip') is not None:  # type: ignore
        raise LoggedError(log, "'thin' and 'skip' should be "
                               "parameters of the 'post' block")

    if sample:
        # If MPI, assume for each MPI process post is passed in the list of
        # collections that should be processed by that process
        # (e.g. single chain output from sampler)
        if isinstance(sample, SampleCollection):
            in_collections = [sample]
        else:
            in_collections = sample
        for i, collection in enumerate(in_collections):
            if skip:
                collection = collection.skip_samples(skip)
            if thin != 1:
                collection = collection.thin_samples(thin)
            in_collections[i] = collection
    elif output_in:
        files = output_in.find_collections()
        numbered = files
        if not numbered:
            # look for un-numbered output files
            files = output_in.find_collections(name=False)
        if files:
            if mpi.size() > len(files):
                raise LoggedError(log, "Number of MPI processes (%s) is larger than "
                                       "the number of sample files (%s)",
                                  mpi.size(), len(files))
            for num in range(mpi.rank(), len(files), mpi.size()):
                in_collections += [SampleCollection(
                    dummy_model_in, output_in,
                    onload_thin=thin, onload_skip=skip, load=True, file_name=files[num],
                    name=str(num + 1) if numbered else "")]
        else:
            raise LoggedError(log, "No samples found for the input model with prefix %s",
                              os.path.join(output_in.folder, output_in.prefix))

    else:
        raise LoggedError(log, "No output from where to load from, "
                               "nor input collections given.")
    # A note on tempered chains: detempering happens automatically when reweighting,
    # which is done later in this function in most cases.
    # But for the sake of robustness, we detemper all chains at init.
    if mpi.is_main_process() and any(c.is_tempered for c in in_collections):
        log.info("Starting from tempered chains. Will detemper before proceeding.")
    # Let's make sure we work on a copy if the chain is going to be altered
    already_copied = bool(output_in) or (sample is not None and (skip or thin != 1))
    for i, collection in enumerate(in_collections):
        if not already_copied:
            collection = collection.copy()
        collection.reset_temperature()
        in_collections[i] = collection
    if any(len(c) <= 1 for c in in_collections):
        raise LoggedError(
            log, "Not enough samples for post-processing. Try using a larger sample, "
                 "or skipping or thinning less.")
    mpi.sync_processes()
    log.info("Will process %d sample points.", sum(len(c) for c in in_collections))

    # 2. Compare old and new info: determine what to do
    add = info_post.get("add") or {}
    if "remove" in add:
        raise LoggedError(log, "remove block should be under 'post', not 'add'")
    remove = info_post.get("remove") or {}
    # Add a dummy 'one' likelihood, to absorb unused parameters
    if not add.get("likelihood"):
        add["likelihood"] = {}
    add["likelihood"]["one"] = None
    # Expand the "add" info, but don't add new default sampled parameters
    orig_params = set(add.get("params") or [])
    add = update_info(add, add_aggr_chi2=False)
    add_params: ExpandedParamsDict = add["params"]  # type: ignore
    for p in set(add_params) - orig_params:
        if p in params_in:
            add_params.pop(p)

    # 2.1 Adding/removing derived parameters and changes in priors of sampled parameters
    out_combined_params = deepcopy_where_possible(params_in)
    remove_params = list(str_to_list(remove.get("params")) or [])
    for p in remove_params:
        pinfo = params_in.get(p)
        if pinfo is None or not is_derived_param(pinfo):
            raise LoggedError(
                log,
                "You tried to remove parameter '%s', which is not a derived parameter. "
                "Only derived parameters can be removed during post-processing.", p)
        out_combined_params.pop(p)
    # Force recomputation of aggregated chi2
    for p in list(out_combined_params):
        if p.startswith(get_chi2_name("")):
            out_combined_params.pop(p)
    prior_recompute_1d = False
    for p, pinfo in add_params.items():
        pinfo_in = params_in.get(p)
        if is_sampled_param(pinfo):
            if not is_sampled_param(pinfo_in):
                # No added sampled parameters (de-marginalisation not implemented)
                if pinfo_in is None:
                    raise LoggedError(
                        log, "You added a new sampled parameter %r (maybe accidentally "
                             "by adding a new likelihood that depends on it). "
                             "Adding new sampled parameters is not possible. Try fixing "
                             "it to some value.", p)
                else:
                    raise LoggedError(
                        log,
                        "You tried to change the prior of parameter '%s', "
                        "but it was not a sampled parameter. "
                        "To change that prior, you need to define as an external one.", p)
            # recompute prior if potentially changed sampled parameter priors
            prior_recompute_1d = True
        elif is_derived_param(pinfo):
            if p in out_combined_params:
                raise LoggedError(
                    log, "You tried to add derived parameter '%s', which is already "
                         "present. To force its recomputation, 'remove' it too.", p)
        elif is_fixed_or_function_param(pinfo):
            # Only one possibility left "fixed" parameter that was not present before:
            # input of new likelihood, or just an argument for dynamical derived (dropped)
            if pinfo_in and p in params_in and pinfo["value"] != pinfo_in.get("value"):
                raise LoggedError(
                    log,
                    "You tried to add a fixed parameter '%s: %r' that was already present"
                    " but had a different value or was not fixed. This is not allowed. "
                    "The old info of the parameter was '%s: %r'",
                    p, dict(pinfo), p, dict(pinfo_in))
        elif not pinfo_in:  # OK as long as we have known value for it
            raise LoggedError(log, "Parameter %s no known value. ", p)
        out_combined_params[p] = pinfo

    out_combined: InputDict = {"params": out_combined_params}  # type: ignore
    # Turn the rest of *derived* parameters into constants,
    # so that the likelihoods do not try to recompute them
    # But be careful to exclude *input* params that have a "derived: True" value
    # (which in "updated info" turns into "derived: 'lambda [x]: [x]'")
    # Don't assign derived parameters to theories, only likelihoods, so they can be
    # recomputed if needed. If the theory does not need to be computed, it doesn't matter
    # if it is already assigned parameters in the usual way; likelihoods can get
    # the required derived parameters from the stored sample derived parameter inputs.
    out_params_with_computed = deepcopy_where_possible(out_combined_params)

    dropped_theory = set()
    for p, pinfo in out_params_with_computed.items():
        if (is_derived_param(pinfo) and "value" not in pinfo
                and p not in add_params):
            out_params_with_computed[p] = {"value": np.nan}
            dropped_theory.add(p)
    # 2.2 Manage adding/removing priors and likelihoods
    warn_remove = False
    kind: ModelBlock
    for kind in ("prior", "likelihood", "theory"):
        out_combined[kind] = deepcopy_where_possible(info_in.get(kind)) or {}
        for remove_item in str_to_list(remove.get(kind)) or []:
            try:
                out_combined[kind].pop(remove_item, None)
                if remove_item not in (add.get(kind) or []) and kind != "theory":
                    warn_remove = True
            except ValueError:
                raise LoggedError(
                    log, "Trying to remove %s '%s', but it is not present. "
                         "Existing ones: %r", kind, remove_item, list(out_combined[kind]))
        if kind != "theory" and kind in add:
            dups = set(add.get(kind) or []).intersection(out_combined[kind]) - {"one"}
            if dups:
                raise LoggedError(
                    log, "You have added %s '%s', which was already present. If you "
                         "want to force its recomputation, you must also 'remove' it.",
                    kind, dups)
            out_combined[kind].update(add[kind])

    if warn_remove and mpi.is_main_process():
        log.warning("You are removing a prior or likelihood pdf. "
                    "Notice that if the resulting posterior is much wider "
                    "than the original one, or displaced enough, "
                    "it is probably safer to explore it directly.")

    mlprior_names_add = minuslogprior_names(add.get("prior") or [])

    out_combined["likelihood"].pop("one", None)

    add_theory = add.get("theory")
    if add_theory:
        if len(add["likelihood"]) == 1 and not any(
                is_derived_param(pinfo) for pinfo in add_params.values()):
            log.warning("You are adding a theory, but this does not force recomputation "
                        "of any likelihood or derived parameters unless explicitly "
                        "removed+added.")
        # Inherit from the original chain (input|output_params, renames, etc)
        added_theory = add_theory.copy()
        for theory, theory_info in out_combined["theory"].items():
            if theory in list(added_theory):
                out_combined["theory"][theory] = \
                    recursive_update(theory_info, added_theory.pop(theory))
        out_combined["theory"].update(added_theory)

    # Prepare recomputation of aggregated chi2
    # (they need to be recomputed by hand, because auto-computation won't pick up
    #  old likelihoods for a given type)
    all_types = {like: str_to_list(opts.get("type") or [])
                 for like, opts in out_combined["likelihood"].items()}
    types = set(chain(*all_types.values()))
    inv_types = {t: [like for like, like_types in all_types.items() if t in like_types]
                 for t in sorted(types)}
    add_aggregated_chi2_params(out_combined_params, types)

    # 3. Create output collection
    # Use default prefix if it exists. If it does not, produce no output by default.
    # {post: {output: None}} suppresses output, and if it's a string, updates it.
    out_prefix = info_post.get("output", info.get("output"))
    if out_prefix:
        suffix = info_post.get("suffix")
        if not suffix:
            raise LoggedError(log, "You need to provide a '%s' for your output chains.",
                              "suffix")
        out_prefix += separator_files + "post" + separator_files + suffix
    output_out = get_output(prefix=out_prefix, force=info.get("force"))
    output_out.set_lock()

    if output_out and not output_out.force and output_out.find_collections():
        raise LoggedError(log, "Found existing post-processing output with prefix %r. "
                               "Delete it manually or re-run with `force: True` "
                               "(or `-f`, `--force` from the shell).", out_prefix)
    elif output_out and output_out.force and mpi.is_main_process():
        output_out.delete_infos()
        for _file in output_out.find_collections():
            output_out.delete_file_or_folder(_file)
    info_out = deepcopy_where_possible(info)
    info_post = info_post.copy()
    info_out["post"] = info_post
    # Updated with input info and extended (updated) add info
    info_out.update(info_in)  # type: ignore
    info_post["add"] = add

    dummy_model_out = DummyModel(out_combined_params, out_combined["likelihood"],
                                 info_prior=out_combined["prior"])
    out_func_parameterization = Parameterization(out_params_with_computed)

    # TODO: check allow_renames=False?
    model_add = Model(out_params_with_computed, add["likelihood"],
                      info_prior=add.get("prior"), info_theory=out_combined["theory"],
                      packages_path=(info_post.get(packages_path_input) or
                                     info.get(packages_path_input)),
                      allow_renames=False, post=True,
                      stop_at_error=info.get('stop_at_error', False),
                      skip_unused_theories=True, dropped_theory_params=dropped_theory)
    # Remove auxiliary "one" before dumping -- 'add' *is* info_out["post"]["add"]
    add["likelihood"].pop("one")
    out_collections = [SampleCollection(dummy_model_out, output_out, name=c.name,
                                        cache_size=OutputOptions.default_post_cache_size)
                       for c in in_collections]
    # TODO: should maybe add skip/thin to out_combined, so can tell post-processed?
    output_out.check_and_dump_info(info_out, out_combined, check_compatible=False)
    collection_in = in_collections[0]
    collection_out = out_collections[0]

    last_percent = None
    known_constants = dummy_model_out.parameterization.constant_params()
    known_constants.update(dummy_model_in.parameterization.constant_params())
    missing_params = dummy_model_in.parameterization.sampled_params().keys() - set(
        collection_in.columns)
    if missing_params:
        raise LoggedError(log, "Input samples do not contain expected sampled parameter "
                               "values: %s", missing_params)

    missing_priors = set(name for name in collection_out.minuslogprior_names if
                         name not in mlprior_names_add
                         and name not in collection_in.columns)
    if _minuslogprior_1d_name in missing_priors:
        prior_recompute_1d = True
    if prior_recompute_1d:
        missing_priors.discard(_minuslogprior_1d_name)
        mlprior_names_add.insert(0, _minuslogprior_1d_name)
    prior_regenerate: Optional[Prior]
    if missing_priors and "prior" in info_in:
        # in case there are input priors that are not stored in input samples
        # e.g. when postprocessing GetDist/CosmoMC-format chains
        in_names = minuslogprior_names(info_in["prior"])
        info_prior = {piname: inf for (piname, inf), in_name in
                      zip(info_in["prior"].items(), in_names) if
                      in_name in missing_priors}
        regenerated_prior_names = minuslogprior_names(info_prior)
        missing_priors.difference_update(regenerated_prior_names)
        prior_regenerate = Prior(dummy_model_in.parameterization, info_prior)
    else:
        prior_regenerate = None
        regenerated_prior_names = None
    if missing_priors:
        raise LoggedError(log, "Missing priors: %s", missing_priors)

    mpi.sync_processes()
    output_in.check_lock()

    # 4. Main loop! Loop over input samples and adjust as required.
    if mpi.is_main_process():
        log.info("Running post-processing...")
    difflogmax: Optional[float] = None
    to_do = sum(len(c) for c in in_collections)
    weights = []
    done = 0
    last_dump_time = time.time()
    for collection_in, collection_out in zip(in_collections, out_collections):
        importance_weights = []

        def set_difflogmax():
            nonlocal difflogmax
            difflog = (collection_in[OutPar.minuslogpost].to_numpy(
                dtype=np.float64)[:len(collection_out)]
                       - collection_out[OutPar.minuslogpost].to_numpy(dtype=np.float64))
            difflogmax = np.max(difflog)
            if abs(difflogmax) < 1:
                difflogmax = 0  # keep simple when e.g. very similar
            log.debug("difflogmax: %g", difflogmax)
            if mpi.more_than_one_process():
                difflogmax = max(mpi.allgather(difflogmax))
            if mpi.is_main_process():
                log.debug("Set difflogmax: %g", difflogmax)
            _weights = np.exp(difflog - difflogmax)
            importance_weights.extend(_weights)
            collection_out.reweight(_weights)

        sampled_params = dummy_model_in.parameterization.sampled_params()

        for i, point in collection_in.data.iterrows():
            all_params = point.to_dict()
            for p in remove_params:
                all_params.pop(p, None)
            log.debug("Point: %r", point)
            sampled = np.array([all_params[param] for param in sampled_params])
            all_params = out_func_parameterization.to_input(all_params).copy()

            # Add/remove priors
            if prior_recompute_1d:
                priors_add = [model_add.prior.logps_internal(sampled)]
                if priors_add[0] == -np.inf:
                    continue
            else:
                priors_add = []
            if model_add.prior.external:
                priors_add.extend(model_add.prior.logps_external(all_params))

            logpriors_add = dict(zip(mlprior_names_add, priors_add))
            logpriors_new = [logpriors_add.get(name, - point.get(name, 0))
                             for name in collection_out.minuslogprior_names]
            if prior_regenerate:
                regenerated = dict(zip(regenerated_prior_names,
                                       prior_regenerate.logps_external(all_params)))
                for _i, name in enumerate(collection_out.minuslogprior_names):
                    if name in regenerated_prior_names:
                        logpriors_new[_i] = regenerated[name]

            if is_debug(log):
                log.debug("New set of priors: %r",
                          dict(zip(dummy_model_out.prior, logpriors_new)))
            if -np.inf in logpriors_new:
                continue
            # Add/remove likelihoods and/or (re-)calculate derived parameters
            loglikes_add, output_derived = model_add._loglikes_input_params(
                all_params, return_output_params=True, as_dict=True)
            loglikes_add = {get_chi2_name(name): loglikes_add[name] for name in
                            model_add.likelihood if name != "one"}
            output_derived = {_p: output_derived[_p] for _p in
                              model_add.output_params}

            loglikes_new = [loglikes_add.get(name, -0.5 * point.get(name, 0))
                            for name in collection_out.chi2_names]
            if is_debug(log):
                log.debug("New set of likelihoods: %r",
                          dict(zip(dummy_model_out.likelihood, loglikes_new)))
                if output_derived:
                    log.debug("New set of derived parameters: %r", output_derived)
            if -np.inf in loglikes_new:
                continue
            all_params.update(output_derived)

            all_params.update(out_func_parameterization.to_derived(all_params))
            derived = {param: all_params.get(param) for param in
                       dummy_model_out.parameterization.derived_params()}
            # We need to recompute the aggregated chi2 by hand
            for type_, likes in inv_types.items():
                derived[get_chi2_name(type_)] = sum(
                    -2 * lvalue for lname, lvalue
                    in zip(collection_out.chi2_names, loglikes_new)
                    if undo_chi2_name(lname) in likes)
            if is_debug(log):
                log.debug("New derived parameters: %r",
                          {p: derived[p]
                           for p in dummy_model_out.parameterization.derived_params()
                           if p in add["params"]})
            # Save to the collection (keep old weight for now)
            weight = point.get(OutPar.weight)
            mpi.check_errors()
            if difflogmax is None and i > OutputOptions.reweight_after and \
                    time.time() - last_dump_time > OutputOptions.output_inteveral_s / 2:
                set_difflogmax()
                collection_out.out_update()

            if difflogmax is not None:
                logpost_new = sum(logpriors_new) + sum(loglikes_new)
                importance_weight = np.exp(logpost_new + point.get(OutPar.minuslogpost)
                                           - difflogmax)
                weight = weight * importance_weight
                importance_weights.append(importance_weight)
                if time.time() - last_dump_time > OutputOptions.output_inteveral_s:
                    collection_out.out_update()
                    last_dump_time = time.time()

            if weight > 0:
                collection_out.add(sampled, derived=derived.values(), weight=weight,
                                   logpriors=logpriors_new, loglikes=loglikes_new)

            # Display progress
            percent = int(np.round((i + done) / to_do * 100))
            if percent != last_percent and not percent % 5:
                last_percent = percent
                progress_bar(log, percent, " (%d/%d)" % (i + done, to_do))

        if difflogmax is None:
            set_difflogmax()
        if not collection_out.data.last_valid_index():
            raise LoggedError(
                log, "No elements in the final sample. Possible causes: "
                     "added a prior or likelihood valued zero over the full sampled "
                     "domain, or the computation of the theory failed everywhere, etc.")
        collection_out.out_update()
        weights.append(np.array(importance_weights))
        done += len(collection_in)

    assert difflogmax is not None
    points = 0
    tot_weight = 0
    min_weight = np.inf
    max_weight = -np.inf
    max_output_weight = -np.inf
    sum_w2 = 0
    points_removed = 0
    for collection_in, collection_out, importance_weights in zip(in_collections,
                                                                 out_collections,
                                                                 weights):
        output_weights = collection_out[OutPar.weight]
        points += len(collection_out)
        tot_weight += np.sum(output_weights)
        points_removed += len(importance_weights) - len(output_weights)
        min_weight = min(min_weight, np.min(importance_weights))
        max_weight = max(max_weight, np.max(importance_weights))
        max_output_weight = max(max_output_weight, np.max(output_weights))
        sum_w2 += np.dot(output_weights, output_weights)

    (tot_weights, min_weights, max_weights, max_output_weights, sum_w2s, points_s,
     points_removed_s) = mpi.zip_gather(
        [tot_weight, min_weight, max_weight, max_output_weight, sum_w2,
         points, points_removed])

    if mpi.is_main_process():
        output_out.clear_lock()
        log.info("Finished! Final number of distinct sample points: %s", sum(points_s))
        log.info("Importance weight range: %.4g -- %.4g",
                 min(min_weights), max(max_weights))
        if sum(points_removed_s):
            log.info("Points deleted due to zero weight: %s", sum(points_removed_s))
        log.info("Effective number of single samples if independent (sum w)/max(w): %s",
                 int(sum(tot_weights) / max(max_output_weights)))
        log.info(
            "Effective number of weighted samples if independent (sum w)^2/sum(w^2): "
            "%s", int(sum(tot_weights) ** 2 / sum(sum_w2s)))
    products: PostResultDict = {"sample": value_or_list(out_collections),
                                "stats": {'min_importance_weight': (min(min_weights) /
                                                                    max(max_weights)),
                                          'points_removed': sum(points_removed_s),
                                          'tot_weight': sum(tot_weights),
                                          'max_weight': max(max_output_weights),
                                          'sum_w2': sum(sum_w2s),
                                          'points': sum(points_s)},
                                "logpost_weight_offset": difflogmax,
                                "weights": value_or_list(weights)}
    return out_combined, products
